from __future__ import absolute_import
import os
import shutil
import unittest
import logging
import tempfile

from six import StringIO


class Base(unittest.TestCase):

    def capture_logging(self, name="", level=logging.INFO,
                        log_file=None, formatter=None):
        if log_file is None:
            log_file = StringIO()
        log_handler = logging.StreamHandler(log_file)
        if formatter:
            log_handler.setFormatter(formatter)
        logger = logging.getLogger(name)
        logger.addHandler(log_handler)
        old_logger_level = logger.level
        logger.setLevel(level)

        @self.addCleanup
        def reset_logging():
            logger.removeHandler(log_handler)
            logger.setLevel(old_logger_level)
        return log_file

    def mkdir(self):
        d = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, d)
        return d

    def change_environment(self, **kw):
        """
        """
        original_environ = dict(os.environ)

        @self.addCleanup
        def cleanup_env():
            os.environ.clear()
            os.environ.update(original_environ)

        os.environ.update(kw)



from __future__ import absolute_import
import subprocess
import os
import re

from .utils import ErrorExit


class Vcs(object):

    err_update = (
        "Could not update branch %(path)s from %(branch_url)s\n\n %(output)s")
    err_branch = "Could not branch %(branch_url)s to %(path)s\n\n %(output)s"
    err_is_mod = "Couldn't determine if %(path)s was modified\n\n %(output)s"
    err_pull = (
        "Could not pull branch @ %(branch_url)s to %(path)s\n\n %(output)s")
    err_cur_rev = (
        "Could not determine current revision %(path)s\n\n %(output)s")

    def __init__(self, path, origin, log):
        self.path = path
        self.log = log
        self.extended_options = self.get_extended_options(origin)

        if self.extended_options:
            self.origin = origin.split("#")[0]
        else:
            self.origin = origin

    def _call(self, args, error_msg, cwd=None, stderr=()):
        try:
            if stderr is not None and not stderr:
                stderr = subprocess.STDOUT
            output = subprocess.check_output(
                args, cwd=cwd or self.path, stderr=stderr)
        except subprocess.CalledProcessError as e:
            self.log.error(error_msg % self.get_err_msg_ctx(e))
            raise ErrorExit()
        return output.strip().decode()

    def get_err_msg_ctx(self, e):
        return {
            'path': self.path,
            'branch_url': self.origin,
            'exit_code': e.returncode,
            'output': e.output,
            'vcs': self.__class__.__name__.lower()}

    def get_extended_options(self, origin):
        regexp = re.compile(r"[\?#&](?P<name>[^&=]+)=(?P<value>[^&=]+)")
        matched = regexp.findall(origin)

        if matched:
            ret = dict()
            for option in matched:
                (name, value) = option
                if name in ret:
                    raise Exception("%s option already defined" % name)
                ret[name] = value
            return ret
        return {}

    def get_cur_rev(self):
        raise NotImplementedError()

    def update(self, rev=None):
        raise NotImplementedError()

    def branch(self):
        raise NotImplementedError()

    def pull(self):
        raise NotImplementedError()

    def is_modified(self):
        raise NotImplementedError()

    # upstream missing revisions?


class Bzr(Vcs):

    def get_cur_rev(self):
        params = ["bzr", "revno", "--tree"]
        return self._call(params, self.err_cur_rev, stderr=None)

    def update(self, rev=None):
        params = ["bzr", "up"]
        if rev:
            params.extend(["-r", str(rev)])
        self._call(params, self.err_update)

    def branch(self):
        params = ["bzr", "co", "--lightweight", self.origin, self.path]
        cwd = os.path.dirname(os.path.dirname(self.path))
        if not cwd:
            cwd = "."
        self._call(params, self.err_branch, cwd)

    def is_modified(self):
        return subprocess.call(
            ["bzr", "diff"],
            cwd=self.path, stdout=subprocess.PIPE) != 0


class Git(Vcs):

    def get_cur_rev(self):
        params = ["git", "rev-parse", "HEAD"]
        return self._call(params, self.err_cur_rev)

    def update(self, rev=None):
        params = ["git", "reset", "--merge"]
        if rev:
            params.append(rev)
        self._call(params, self.err_update)

    def branch(self):
        params = ["git", "clone", "--depth", "1"]
        # Deal with branches in the format
        # <repository-url>;<branch>
        components = self.origin.split(';')
        if len(components) == 2:
            params += ["--branch", components[1],
                       components[0], self.path]
        else:
            params += [self.origin, self.path]
        cwd = os.path.dirname(os.path.dirname(self.path))
        if not cwd:
            cwd = "."

        self._call(params, self.err_branch, cwd)

        change_ref = self.extended_options.get('changeref', None)
        if change_ref:
            change_ref = 'refs/changes/{}/{}'.format(
                change_ref.split("/")[0][-2:], change_ref)

            self._call(["git", "fetch", "--depth", "1",
                        self.origin, change_ref],
                       self.err_branch, self.path)
            self._call(["git", "checkout", "FETCH_HEAD"], self.err_branch,
                       self.path)

    def is_modified(self):
        params = ["git", "status", "-s"]
        return bool(self._call(params, self.err_is_mod).strip())

    def get_remote_origin(self):
        params = ["git", "config", "--get", "remote.origin.url"]
        return self._call(params, "")



"""Tests for the GUI server bundles deployment support."""

from __future__ import absolute_import
from contextlib import contextmanager
import os
import shutil
import tempfile
import unittest

import mock
import yaml

from deployer import guiserver
from deployer.feedback import Feedback
from deployer.tests.base import skip_if_offline


class TestGetDefaultGuiserverOptions(unittest.TestCase):

    def setUp(self):
        self.options = guiserver.get_default_guiserver_options()

    def test_option_keys(self):
        # All the required options are returned.
        # When adding/modifying options, ensure the defaults are sane for us.
        expected_keys = set([
            'bootstrap', 'branch_only', 'configs', 'debug', 'deploy_delay',
            'deployment', 'description', 'destroy_applications', 'diff',
            'find_application', 'ignore_errors', 'juju_env', 'list_deploys',
            'no_local_mods', 'no_relations', 'overrides', 'rel_wait',
            'retry_count', 'series', 'skip_unit_wait', 'terminate_machines',
            'timeout', 'update_charms', 'verbose', 'watch'
        ])
        self.assertEqual(expected_keys, set(self.options.__dict__.keys()))

    def test_option_values(self):
        # The options values are suitable to be used by the GUI server.
        # When adding/modifying options, ensure the defaults are sane for us.
        options = self.options
        self.assertFalse(options.bootstrap)
        self.assertFalse(options.branch_only)
        self.assertIsNone(options.configs)
        self.assertFalse(options.debug)
        self.assertEqual(0, options.deploy_delay)
        self.assertIsNone(options.deployment)
        self.assertFalse(options.destroy_applications)
        self.assertFalse(options.diff)
        self.assertIsNone(options.find_application)
        self.assertTrue(options.ignore_errors)
        self.assertEqual(os.getenv("JUJU_ENV"), options.juju_env)
        self.assertFalse(options.list_deploys)
        self.assertTrue(options.no_local_mods)
        self.assertIsNone(options.overrides)
        self.assertEqual(60, options.rel_wait)
        self.assertEqual(0, options.retry_count)
        self.assertIsNone(options.series)
        self.assertFalse(options.terminate_machines)
        self.assertEqual(2700, options.timeout)
        self.assertFalse(options.update_charms)
        self.assertFalse(options.verbose)
        self.assertFalse(options.watch)


class TestDeploymentError(unittest.TestCase):

    def test_error(self):
        # A single error is properly stored and represented.
        exception = guiserver.DeploymentError(['bad wolf'])
        self.assertEqual(['bad wolf'], exception.errors)
        self.assertEqual('bad wolf', str(exception))

    def test_multiple_errors(self):
        # Multiple deployment errors are properly stored and represented.
        errors = ['error 1', 'error 2']
        exception = guiserver.DeploymentError(errors)
        self.assertEqual(errors, exception.errors)
        self.assertEqual('error 1\nerror 2', str(exception))


class TestGUIDeployment(unittest.TestCase):

    def setUp(self):
        # Set up a GUIDeployment instance and a Feedback object.
        self.deployment = guiserver.GUIDeployment('mybundle', 'mydata', 4)
        self.feedback = Feedback()

    def test_valid_deployment(self):
        # If the bundle is well formed, the deployment proceeds normally.
        self.assertIsNone(self.deployment._handle_feedback(self.feedback))

    def test_warnings(self):
        # Warning messages are properly logged.
        self.feedback.warn('we are the Borg')
        with mock.patch.object(self.deployment, 'log') as mock_log:
            self.deployment._handle_feedback(self.feedback)
        mock_log.warning.assert_called_once_with('we are the Borg')

    def test_errors(self):
        # A DeploymentError is raised if errors are found in the bundle.
        self.feedback.error('error 1')
        self.feedback.error('error 2')
        with self.assertRaises(guiserver.DeploymentError) as cm:
            self.deployment._handle_feedback(self.feedback)
        self.assertEqual(['error 1', 'error 2'], cm.exception.errors)


class DeployerFunctionsTestMixin(object):
    """Base set up for the functions that make use of the juju-deployer."""

    apiurl = 'wss://api.example.com:17070'
    username = 'who'
    password = 'Secret!'
    name = 'mybundle'
    bundle = yaml.safe_load("""
      services:
        wordpress:
          charm: "cs:precise/wordpress-20"
          num_units: 1
          options:
            debug: "no"
            engine: nginx
            tuning: single
          annotations:
            "gui-x": "425.347"
            "gui-y": "414.547"
        mysql:
          charm: "cs:precise/mysql-28"
          num_units: 2
          constraints:
            arch: i386
            mem: 4G
            cpu-cores: 4
          annotations:
            "gui-x": "494.347"
            "gui-y": "164.547"
      relations:
        - - "mysql:db"
          - "wordpress:db"
      series: precise
    """)

    def check_environment_life(self, mock_environment):
        """Check the calls executed on the given mock environment.

        Ensure that, in order to retrieve the list of currently deployed
        services, the environment is instantiated, connected, env.status is
        called and then the connection is closed.
        """
        mock_environment.assert_called_once_with(
            self.apiurl, self.username, self.password)
        mock_env_instance = mock_environment()
        mock_env_instance.connect.assert_called_once_with()
        mock_env_instance.status.assert_called_once_with()
        mock_env_instance.close.assert_called_once_with()

    @contextmanager
    def assert_overlapping_services(self, mock_environment):
        """Ensure a ValueError is raised in the context manager block.

        The given mock environment object is set up so that its status
        simulates an existing service. The name of this service overlaps with
        the name of one of the services in the bundle.
        """
        mock_env_instance = mock_environment()
        mock_env_instance.status.return_value = {'services': {'mysql': {}}}
        # Ensure a ValueError is raised by the code in the context block.
        with self.assertRaises(ValueError) as context_manager:
            yield
        # The error reflects the overlapping service name.
        error = str(context_manager.exception)
        self.assertEqual('service(s) already in the environment: mysql', error)
        # Even if an error occurs, the environment connection is closed.
        mock_env_instance.close.assert_called_once_with()


@skip_if_offline
@mock.patch('deployer.guiserver.GUIEnvironment')
class TestValidate(DeployerFunctionsTestMixin, unittest.TestCase):

    def test_validation(self, mock_environment):
        # The validation is correctly run.
        guiserver.validate(
            self.apiurl, self.username, self.password, self.bundle)
        # The environment is correctly instantiated and used.
        self.check_environment_life(mock_environment)

    def test_overlapping_services(self, mock_environment):
        # The validation fails if the bundle includes a service name already
        # present in the Juju environment.
        with self.assert_overlapping_services(mock_environment):
            guiserver.validate(
                self.apiurl, self.username, self.password, self.bundle)


@skip_if_offline
@mock.patch('deployer.guiserver.GUIEnvironment')
class TestImportBundle(DeployerFunctionsTestMixin, unittest.TestCase):

    # The options attribute simulates the options passed to the Importer.
    options = guiserver.get_default_guiserver_options()

    @contextmanager
    def patch_juju_home(self):
        """Patch the value used by the bundle importer as Juju home."""
        base_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, base_dir)
        juju_home = os.path.join(base_dir, 'juju-home')
        with mock.patch('deployer.guiserver.JUJU_HOME', juju_home):
            try:
                yield juju_home
            finally:
                del os.environ['JUJU_HOME']

    def import_bundle(self, version=4):
        """Call the import_bundle function."""
        guiserver.import_bundle(
            self.apiurl, self.username, self.password, self.name, self.bundle,
            version, self.options)

    def cleanup_series_path(self):
        """Remove the series path created by the Deployment object."""
        if os.path.isdir('precise'):
            os.rmdir('precise')

    @mock.patch('deployer.guiserver.Importer')
    def test_importing_bundle(self, mock_importer, mock_environment):
        # The juju-deployer importer is correctly set up and run.
        with self.patch_juju_home():
            self.import_bundle()
        # The environment is correctly instantiated and used.
        self.check_environment_life(mock_environment)
        # The importer is correctly instantiated.
        self.assertEqual(1, mock_importer.call_count)
        importer_args = mock_importer.call_args[0]
        self.assertEqual(3, len(importer_args))
        env, deployment, options = importer_args
        # The first argument passed to the importer is the environment.
        self.assertIs(mock_environment(), env)
        # The second argument is the deployment object.
        self.assertIsInstance(deployment, guiserver.GUIDeployment)
        self.assertEqual(self.name, deployment.name)
        self.assertEqual(self.bundle, deployment.data)
        # The third and last argument is the options object.
        self.assertIs(self.options, options)
        # The importer is started.
        mock_importer().run.assert_called_once_with()

    def test_overlapping_services(self, mock_environment):
        # The import fails if the bundle includes a service name already
        # present in the Juju environment.
        with self.assert_overlapping_services(mock_environment):
            with self.patch_juju_home():
                self.import_bundle()

    @mock.patch('deployer.guiserver.Importer')
    def test_juju_home(self, mock_importer, mock_environment):
        # A customized Juju home is created and used during the import process.
        with self.patch_juju_home() as juju_home:
            assert not os.path.isdir(juju_home), 'directory should not exist'

            def run():
                # Ensure JUJU_HOME is included in the context when the Importer
                # instance is run.
                self.assertEqual(juju_home, os.getenv('JUJU_HOME'))
            mock_importer().run = run
            self.import_bundle()
        # The JUJU_HOME directory has been created.
        self.assertTrue(os.path.isdir(juju_home))

    @mock.patch('time.sleep')
    def test_importer_behavior(self, mock_sleep, mock_environment):
        # The importer executes the expected environment calls.
        self.addCleanup(self.cleanup_series_path)
        with self.patch_juju_home():
            self.import_bundle(version=3)
        mock_sleep.assert_has_calls([mock.call(5.1), mock.call(60)])
        # If any of the calls below fails, then we have to change the
        # signatures of deployer.guiserver.GUIEnvironment methods.
        mock_environment.assert_called_once_with(
            self.apiurl, self.username, self.password)
        mock_environment().assert_has_calls([
            mock.call.connect(),
            mock.call.status(),
            mock.call.deploy(
                'mysql', 'cs:precise/mysql-28', '', None,
                {'arch': 'i386', 'cpu-cores': 4, 'mem': '4G'},
                None, None, 2, None, 'precise', None),
            mock.call.set_annotation(
                'mysql', {'gui-y': '164.547', 'gui-x': '494.347'}),
            mock.call.deploy(
                'wordpress', 'cs:precise/wordpress-20', '',
                {'debug': 'no', 'engine': 'nginx', 'tuning': 'single'},
                None, None, None, 1, None, 'precise', None),
            mock.call.set_annotation(
                'wordpress', {'gui-y': '414.547', 'gui-x': '425.347'}),
            mock.call.add_units('mysql', 2),
            mock.call.add_units('wordpress', 1),
            mock.call.add_relation('mysql:db', 'wordpress:db'),
            mock.call.close(),
        ], any_order=True)

    def test_deployment_errors(self, mock_environment):
        # A DeploymentError is raised if the deployment fails.
        bundle = {
            'services': {
                'wordpress': {
                    'charm': 'cs:precise/wordpress-20',
                    'options': {'no-such-option': 42},  # Invalid options.
                },
                'mysql': {
                    'charm': 'cs:precise/mysql-28',
                    'options': {'bad': 'wolf'},  # Invalid options.
                },
            },
        }
        version = 4
        self.addCleanup(self.cleanup_series_path)
        with self.patch_juju_home():
            with self.assertRaises(guiserver.DeploymentError) as cm:
                guiserver.import_bundle(
                    self.apiurl, self.username, self.password, self.name,
                    bundle, version, self.options)
        expected_errors = set([
            'Invalid config charm cs:precise/wordpress-20 no-such-option=42',
            'Invalid config charm cs:precise/mysql-28 bad=wolf',
        ])
        self.assertEqual(expected_errors, set(cm.exception.errors))



from __future__ import absolute_import
import logging
import time

from .base import BaseAction
from ..env import watchers
from ..utils import ErrorExit
from six.moves import map
from six.moves import range


class Importer(BaseAction):

    log = logging.getLogger("deployer.import")

    def __init__(self, env, deployment, options):
        self.options = options
        self.env = env
        self.deployment = deployment

    def add_units(self):
        self.log.debug("Adding units...")
        # Add units to existing services that don't match count.
        env_status = self.env.status()

        for svc in self.deployment.get_services():
            charm = self.deployment.get_charm_for(svc.name)
            if charm.is_subordinate():
                self.log.warning(
                    "Config specifies num units for subordinate: %s",
                    svc.name)
                continue

            cur_units = len(
                env_status['services'][svc.name].get('units') or ())
            delta = (svc.num_units - cur_units)

            if delta <= 0:
                self.log.debug(
                    " Application %r does not need any more units added.",
                    svc.name)
                continue

            self.log.info(
                "Adding %d more units to %s" % (abs(delta), svc.name))
            if svc.unit_placement:
                # Reload status before each placed unit is deployed, so that
                # co-location to other units can take place properly.
                env_status = self.env.status()
                placement = self.deployment.get_unit_placement(svc, env_status)
                for mid in range(cur_units, svc.num_units):
                    self.env.add_unit(svc.name, placement.get(mid))
            else:
                self.env.add_units(svc.name, abs(delta))

    def machine_exists(self, id):
        """Checks if the given id exists on the current environment."""
        return str(id) in list(map(str, self.env.status().get('machines', {})))

    def create_machines(self):
        """Create machines as specified in the machine spec in the bundle.

        A machine spec consists of a named machine (the name is, by convention,
        an integer) with an optional series, optional constraints and optional
        annotations:

        0:
          series: "precise"
          constraints: "mem=4G arch=i386"
          annotations:
            foo: bar
        1:
          constraints: "mem=4G"

        This method first attempts to create any machines in the 'machines'
        section of the bundle specification with the given constraints and
        annotations.  Then, if there are any machines requested for containers
        in the style of <container type>:new, it creates those machines and
        adds them to the machines map.

        If the machine already exists on the environment, then the new machine
        creation is skipped.
        """
        machines = self.deployment.get_machines()
        machines_map = {}

        if machines:
            self.log.info("Creating machines...")
            for machine_name, spec in sorted(machines.items()):
                if self.machine_exists(machine_name):
                    # XXX frankban: do we really want this? The machine
                    # identifiers as included in v4 bundles are not intended
                    # to refer to real machines. A mapping could be provided
                    # but this kind of implicit mapping seems weak.
                    self.log.info(
                        """Machine %s already exists on environment, """
                        """skipping creation""" % machine_name)
                    machines_map[machine_name] = str(machine_name)
                else:
                    self.log.info("Machine %s will be created" % machine_name)
                    machines_map[machine_name] = self.env.add_machine(
                        series=spec.get('series',
                                        self.deployment.data['series']),
                        constraints=spec.get('constraints'))

                if isinstance(spec, dict):
                    annotations = spec.get('annotations')
                    if annotations:
                        self.env.set_annotation(
                            machines_map[machine_name],
                            annotations,
                            entity_type='machine')

            # In the case of <container type>:new, we need to create a machine
            # before creating the container on which the service will be
            # deployed.  This is stored in the machines map which will be used
            # in the service placement.
            for service in self.deployment.get_services():
                placement = self.deployment.get_unit_placement(service, None)
                for container_host in \
                        placement.get_new_machines_for_containers():
                    if self.machine_exists(machine_name):
                        self.log.info("Machine %s already exists,"
                                      "skipping creation" % machine_name)
                        machines_map[container_host] = str(container_host)
                    else:
                        self.log.info("A new container will be created"
                                      "on machine: %s" % container_host)
                        machines_map[container_host] = self.env.add_machine()
            self.deployment.set_machines(machines_map)

    def get_charms(self):
        # Get Charms
        self.log.debug("Getting charms...")
        self.deployment.fetch_charms(
            update=self.options.update_charms,
            no_local_mods=self.options.no_local_mods)

        # Load config overrides/includes and verify rels after we can
        # validate them.
        self.deployment.resolve(self.options.overrides or ())

    def deploy_services(self, add_units=True):
        """Deploy the services specified in the deployment.

        add_units: whether or not to add units to the service as it is
          deployed; newer versions of bundles may have machines specified
          in a machine spec, and units will be placed accordingly if this
          flag is false.
        """
        self.log.info("Deploying applications...")
        env_status = self.env.status()
        reloaded = False

        for svc in self.deployment.get_services():
            if svc.name in env_status['services']:
                self.log.debug(
                    " Application %r already deployed. Skipping" % svc.name)
                continue

            charm = self.deployment.get_charm_for(svc.name)
            self.log.info(
                " Deploying application %s using %s", svc.name,
                charm.charm_url if not charm.is_absolute() else charm.path
            )

            if svc.unit_placement:
                # We sorted all the non placed services first, so we only
                # need to update status once after we're done with them, in
                # the instance of v3 bundles; in the more complex case of v4
                # bundles, we'll need to refresh each time.
                if not reloaded:
                    self.log.debug(
                        " Refetching status for placement deploys")
                    time.sleep(5.1)
                    env_status = self.env.status()
                    # In the instance of version 3 deployments, we will not
                    # need to fetch the status more than once.  In version 4
                    # bundles, however, we will need to fetch the status each
                    # time in order to allow for the machine specification to
                    # be taken into account.
                    if self.deployment.version == 3:
                        reloaded = True
                num_units = 1
            else:
                num_units = svc.num_units

            # Only add a single unit if requested. This is done after the above
            # work to ensure that the status is still retrieved as necessary.
            if not add_units:
                num_units = 1

            placement = self.deployment.get_unit_placement(svc, env_status)

            if charm.is_subordinate():
                num_units = None

            self.env.deploy(
                svc.name,
                charm.charm_url,
                charm.repo_path or self.deployment.repo_path,
                svc.config,
                svc.constraints,
                svc.resources,
                svc.storage,
                num_units,
                placement.get(0),
                charm.series or self.deployment.series,
                svc.bindings,
            )

            if svc.annotations:
                self.log.debug(" Setting annotations")
                self.env.set_annotation(svc.name, svc.annotations)

            if self.options.deploy_delay:
                self.log.debug(" Waiting for deploy delay")
                time.sleep(self.options.deploy_delay)

    def add_relations(self):
        self.log.info("Adding relations...")

        # Relations
        status = self.env.status()
        created = False

        for end_a, end_b in self.deployment.get_relations():
            if self._rel_exists(status, end_a, end_b):
                continue
            self.log.info(" Adding relation %s <-> %s", end_a, end_b)
            self.env.add_relation(end_a, end_b)
            created = True
        return created

    def _rel_exists(self, status, end_a, end_b):
        # Checks for a named relation on one side that matches the local
        # endpoint and remote service.
        (name_a, name_b, rem_a, rem_b) = (end_a, end_b, None, None)

        if ":" in end_a:
            name_a, rem_a = end_a.split(":", 1)
        if ":" in end_b:
            name_b, rem_b = end_b.split(":", 1)

        rels_svc_a = status['services'][name_a].get('relations', {})

        found = False
        for r, related in rels_svc_a.items():
            if name_b in related:
                if rem_a and r not in rem_a:
                    continue
                found = True
                break
        if found:
            return True
        return False

    def check_timeout(self):
        timeout = self.options.timeout - (time.time() - self.start_time)
        if timeout < 0:
            self.log.error("Reached deployment timeout.. exiting")
            raise ErrorExit()
        return timeout

    def wait_for_units(self, ignore_errors=False):
        if self.options.skip_unit_wait:
            return
        timeout = self.check_timeout()
        # Set up the callback to be called in case of unit errors: if
        # ignore_errors is True errors are just logged, otherwise we exit the
        # program.
        if ignore_errors:
            on_errors = watchers.log_on_errors(self.env)
        else:
            on_errors = watchers.exit_on_errors(self.env)
        self.env.wait_for_units(
            int(timeout), watch=self.options.watch,
            services=self.deployment.get_service_names(), on_errors=on_errors)

    def run(self):
        options = self.options
        self.start_time = time.time()

        # Get charms
        self.get_charms()
        if options.branch_only:
            return

        if options.bootstrap:
            self.env.bootstrap()
        self.env.connect()

        if self.deployment.version > 3:
            self.create_machines()

        # We can shortcut and add the units during deployment for v3 bundles.
        self.deploy_services(add_units=(self.deployment.version == 3))

        # Workaround api issue in juju-core, where any action takes 5s
        # to be consistent to subsequent watch api interactions, see
        # http://pad.lv/1203105 which will obviate this wait.
        time.sleep(5.1)
        self.add_units()

        ignore_errors = bool(options.retry_count) or options.ignore_errors
        self.log.debug("Waiting for units before adding relations")
        self.wait_for_units(ignore_errors=ignore_errors)

        # Reset our environment connection, as it may grow stale during
        # the watch (we're using a sync client so not responding to pings
        # unless actively using the conn).
        self.env.close()
        self.env.connect()

        self.check_timeout()
        rels_created = False
        if not options.no_relations:
            rels_created = self.add_relations()

        # Wait for the units to be up before waiting for rel stability.
        if rels_created and not options.skip_unit_wait:
            self.log.debug(
                "Waiting for relation convergence %ds", options.rel_wait)
            time.sleep(options.rel_wait)
            self.wait_for_units(ignore_errors=ignore_errors)

        if options.retry_count:
            self.log.info("Looking for errors to auto-retry")
            self.env.resolve_errors(
                options.retry_count,
                options.timeout - time.time() - self.start_time)

        # Finally expose things
        for svc in self.deployment.get_services():
            if svc.expose:
                self.log.info(" Exposing application %r" % svc.name)
                self.env.expose(svc.name)



from __future__ import absolute_import
import os
from subprocess import CalledProcessError

from mock import (
    MagicMock,
    patch,
)

from six import BytesIO

from .base import Base
from deployer.utils import (
    _check_call,
    _is_qualified_charm_url,
    DeploymentError,
    dict_merge,
    ErrorExit,
    get_qualified_charm_url,
    HTTPError,
    mkdir,
    URLError,
)


class UtilTests(Base):

    def test_relation_list_merge(self):
        self.assertEqual(
            dict_merge(
                {'relations': [['m1', 'x1']]},
                {'relations': [['m2', 'x2']]}),
            {'relations': [['m1', 'x1'], ['m2', 'x2']]})

    def test_no_rels_in_target(self):
        self.assertEqual(
            dict_merge(
                {'a': 1},
                {'relations': [['m1', 'x1'], ['m2', 'x2']]}),
            {'a': 1, 'relations': [['m1', 'x1'], ['m2', 'x2']]})

    @patch('subprocess.check_output')
    def test_check_call_fail_no_retry(self, check_output):
        _e = CalledProcessError(returncode=1, cmd=['fail'])
        check_output.side_effect = _e
        self.assertRaises(
            ErrorExit, _check_call, params=['fail'], log=MagicMock())

    @patch('time.sleep')
    @patch('subprocess.check_output')
    def test_check_call_fail_retry(self, check_output, sleep):
        _e = CalledProcessError(returncode=1, cmd=['fail'])
        check_output.side_effect = _e
        self.assertRaises(
            ErrorExit, _check_call, params=['fail'], log=MagicMock(),
            max_retry=3)
        # 1 failure + 3 retries
        self.assertEquals(len(check_output.call_args_list), 4)

    @patch('time.sleep')
    @patch('subprocess.check_output')
    def test_check_call_succeed_after_retry(self, check_output, sleep):
        # call succeeds after the 3rd try
        _e = CalledProcessError(returncode=1, cmd=['maybe_fail'])
        check_output.side_effect = [
            _e, _e, 'good', _e]
        output = _check_call(
            params=['maybe_fail'], log=MagicMock(), max_retry=3)
        self.assertEquals(output, 'good')
        # 1 failure + 3 retries
        self.assertEquals(len(check_output.call_args_list), 3)

    def test_check_call_uses_shell(self):
        cmd = 'echo "foo"'
        self.assertRaises(
            OSError, _check_call, params=[cmd], log=MagicMock())
        output = _check_call(params=[cmd], log=MagicMock(), shell=True)
        self.assertEqual(output, b"foo\n")


class TestMkdir(Base):

    def setUp(self):
        self.playground = self.mkdir()

    def test_create_dir(self):
        # A directory is correctly created.
        path = os.path.join(self.playground, 'foo')
        mkdir(path)
        self.assertTrue(os.path.isdir(path))

    def test_intermediate_dirs(self):
        # All intermediate directories are created.
        path = os.path.join(self.playground, 'foo', 'bar', 'leaf')
        mkdir(path)
        self.assertTrue(os.path.isdir(path))

    def test_expand_user(self):
        # The ~ construction is expanded.
        with patch('os.environ', {'HOME': self.playground}):
            mkdir('~/in/my/home')
        path = os.path.join(self.playground, 'in', 'my', 'home')
        self.assertTrue(os.path.isdir(path))

    def test_existing_dir(self):
        # The function exits without errors if the target directory exists.
        path = os.path.join(self.playground, 'foo')
        os.mkdir(path)
        mkdir(path)

    def test_existing_file(self):
        # An OSError is raised if a file already exists in the target path.
        path = os.path.join(self.playground, 'foo')
        with open(path, 'w'):
            with self.assertRaises(OSError):
                mkdir(path)

    def test_failure(self):
        # Errors are correctly re-raised.
        path = os.path.join(self.playground, 'foo')
        os.chmod(self.playground, 0000)
        self.addCleanup(os.chmod, self.playground, 0o700)
        with self.assertRaises(OSError):
            mkdir(os.path.join(path))
        self.assertFalse(os.path.exists(path))


class TestCharmRevisioning(Base):
    """Test the functions related to charm URL revisioning."""

    def test_is_qualified_false(self):
        url = "cs:precise/mysql"
        self.assertFalse(_is_qualified_charm_url(url))

    def test_is_qualified_1_digit(self):
        url = "cs:precise/mysql-2"
        self.assertTrue(_is_qualified_charm_url(url))

    def test_is_qualified_many_digits(self):
        url = "cs:precise/mysql-2014"
        self.assertTrue(_is_qualified_charm_url(url))

    def test_is_qualified_no_digits(self):
        url = "cs:precise/mysql-"
        self.assertFalse(_is_qualified_charm_url(url))

    def test_get_qualified_url(self):
        fake_json = b"""
          {"cs:precise/mysql":
            {"revision":333}
          }
        """

        def mocked_urlopen(url):
            return BytesIO(fake_json)

        path = 'deployer.utils.urlopen'
        with patch(path, mocked_urlopen):
            url = get_qualified_charm_url('cs:precise/mysql')
        self.assertEqual('cs:precise/mysql-333', url)

    def test_get_qualified_url_raise_exception_on_HTTPError(self):
        def mocked_urlopen(url):
            raise HTTPError(url, 404, 'Bad Earl', None, None)

        with patch('deployer.utils.urlopen', mocked_urlopen):
            with self.assertRaises(DeploymentError) as exc:
                get_qualified_charm_url('cs:precise/mysql')
            expected = (
                'HTTP Error 404: '
                'Bad Earl (https://api.jujucharms.com/charmstore/charm-info'
                '?charms=cs:precise/mysql)')
            self.assertEqual([expected], exc.exception.args[0])

    def test_get_qualified_url_raise_exception_on_URLError(self):
        def mocked_urlopen(url):
            raise URLError('Hinky URL')

        with patch('deployer.utils.urlopen', mocked_urlopen):
            with self.assertRaises(DeploymentError) as exc:
                get_qualified_charm_url('cs:precise/mysql')
            expected = ('<urlopen error Hinky URL> '
                        '(https://api.jujucharms.com/charmstore/charm-info'
                        '?charms=cs:precise/mysql)')
            self.assertEqual([expected], exc.exception.args[0])



from __future__ import absolute_import
import logging

from .base import BaseAction


class Export(BaseAction):

    log = logging.getLogger("deployer.export")

    def __init__(self, env, deployment, options):
        self.options = options
        self.env = env
        self.deployment = deployment
        self.env_status = None
        self.env_state = {'services': {}, 'relations': {}}

    def run(self):
        pass



from __future__ import absolute_import
import os
import logging
import subprocess

from deployer.charm import Charm
from deployer.utils import ErrorExit, yaml_dump
from deployer.vcs import Bzr as BaseBzr
from deployer.vcs import Git as BaseGit
from .base import Base, skip_if_offline


class StoreCharmTest(Base):
    def setUp(self):
        self.charm = Charm(
            'cs:trusty/ubuntu', None, None, None, None)

    def test_is_absolute(self):
        self.assertFalse(self.charm.is_absolute())

    def test_repo_path(self):
        self.assertIsNone(self.charm.repo_path)

    def test_series(self):
        def _c(url, deploy_series, data_series):
            return Charm.from_service(url, None, deploy_series, {
                'series': data_series,
            } if data_series else {}).series
        self.assertEqual('trusty', _c('cs:trusty/ubuntu', None, None))
        self.assertEqual('trusty', _c('cs:trusty/ubuntu', 'xenial', None))
        self.assertEqual('xenial', _c('cs:trusty/ubuntu', 'trusty', 'xenial'))
        self.assertEqual('xenial', _c('cs:ubuntu', 'trusty', 'xenial'))
        self.assertEqual('trusty', _c('cs:ubuntu', 'trusty', None))
        self.assertIsNone(_c('cs:ubuntu', None, None))


class AbsolutePathCharmTest(Base):
    def setUp(self):
        self.charm = Charm(
            '/tmp/trusty/ubuntu', '/tmp/trusty/ubuntu', None, None, None)

    def test_is_absolute(self):
        self.assertTrue(self.charm.is_absolute())

    def test_repo_path(self):
        self.assertEqual('/tmp', self.charm.repo_path)


class Bzr(BaseBzr):

    def __init__(self, path):
        super(Bzr, self).__init__(
            path, "", logging.getLogger("deployer.repo"))

    def init(self):
        self._call(
            ["bzr", "init", self.path],
            "Could not initialize repo at  %(path)s")
        try:
            subprocess.check_output(["bzr", "whoami"])
        except subprocess.CalledProcessError:
            subprocess.check_call([
                "bzr", "whoami", "TestRunner <test@example.com>"])

    def write(self, files):
        for f in files:
            with open(os.path.join(
                    self.path, f), 'w') as fh:
                fh.write(files[f])
            self._call(
                ["bzr", "add", f],
                "Could not add file %s" % f)

    def commit(self, msg):
        self._call(
            ["bzr", "commit", "-m", msg],
            "Could not commit at %(path)s")

    def revert(self):
        self._call(
            ["bzr", "revert"],
            "Could not revert at %(path)s")

    def tag(self, name):
        self._call(
            ["bzr", "tag", name],
            "Could not tag at %(path)s")

    branch = update = pull = None


class BzrCharmTest(Base):

    def setUp(self):
        # Isolate tests from the user's bzr config.
        orig_home = os.environ.get('HOME')
        os.environ['HOME'] = self.mkdir()
        self.addCleanup(os.environ.update, {'HOME': orig_home})

        self.repo_path = d = self.mkdir()
        self.series_path = os.path.join(d, "precise")
        os.mkdir(self.series_path)
        self.output = self.capture_logging(
            "deployer.charm", level=logging.DEBUG)

    def setup_vcs_charm(self):
        self.branch = Bzr(self.mkdir())
        self.branch.init()
        self.branch.write(
            {'metadata.yaml': yaml_dump({
                'name': 'couchdb',
                'summary': 'RESTful document oriented database',
                'provides': {
                    'db': {
                        'interface': 'couchdb'}}}),
             'revision': '3'})
        self.branch.commit('initial')
        self.branch.write({'revision': '4'})
        self.branch.commit('next')
        self.branch.tag('v2')
        self.branch.write({'revision': '5'})
        self.branch.commit('next')

        self.charm_data = {
            "charm": "couchdb",
            "build": None,
            "branch": self.branch.path,
            "rev": None,
            "charm_url": None,
        }

    @skip_if_offline
    def test_vcs_charm(self):
        self.setup_vcs_charm()
        params = dict(self.charm_data)
        charm = Charm.from_service(
            "scratch", self.repo_path, "precise", params)

        charm.fetch()
        self.assertEqual(charm.metadata['name'], 'couchdb')
        self.assertEqual(charm.vcs.get_cur_rev(), '3')

        charm.rev = '2'
        charm.update()
        self.assertEqual(charm.vcs.get_cur_rev(), '2')

        self.assertFalse(charm.is_modified())
        with open(os.path.join(charm.path, 'revision'), 'w') as fh:
            fh.write('0')
        self.assertTrue(charm.is_modified())
        Bzr(charm.path).revert()

        charm.rev = None
        # Update goes to latest with no revision specified
        charm.update()
        self.assertEqual(charm.vcs.get_cur_rev(), '3')

    @skip_if_offline
    def test_vcs_fetch_with_rev(self):
        self.setup_vcs_charm()
        params = dict(self.charm_data)
        params['branch'] = params['branch'] + '@2'
        charm = Charm.from_service(
            "scratch", self.repo_path, "precise", params)
        charm.fetch()
        self.assertEqual(charm.vcs.get_cur_rev(), '2')

    charms_vcs_series = [
        ({"charm": "local:precise/mongodb",
          "branch": "lp:charms/precise/couchdb"},
         'trusty', 'precise/mongodb'),
        ({"series": "trusty",
          "charm": "couchdb",
          "branch": "lp:charms/precise/couchdb"},
         'precise', 'trusty/couchdb')]

    def test_vcs_charm_with_series(self):
        for data, dseries, path in self.charms_vcs_series:
            charm = Charm.from_service(
                "db", "/tmp", dseries, data)
            self.assertEqual(
                charm.path,
                os.path.join('/tmp', path))
            self.assertEqual(
                charm.series_path, os.path.join('/tmp', path.split('/')[0]))

    def test_charm_error(self):
        branch = self.mkdir()
        params = {
            'charm': 'couchdb',
            'branch': "file://%s" % branch}
        charm = Charm.from_service(
            "scratch", self.repo_path, "precise", params)
        self.assertRaises(ErrorExit, charm.fetch)
        self.assertIn('bzr: ERROR: Not a branch: ', self.output.getvalue())


class Git(BaseGit):

    def __init__(self, path):
        super(Git, self).__init__(
            path, "", logging.getLogger("deployer.repo"))

    def init(self):
        self._call(
            ["git", "init", self.path],
            "Could not initialize repo at  %(path)s")
        self._call(
            ["git", "config", "user.email", "test@example.com"],
            "Could not config user.email at  %(path)s")
        self._call(
            ["git", "config", "user.name", "TestRunner"],
            "Could not config user.name at  %(path)s")

    def write(self, files):
        for f in files:
            with open(os.path.join(
                    self.path, f), 'w') as fh:
                fh.write(files[f])
            self._call(
                ["git", "add", f],
                "Could not add file %s" % f)

    def commit(self, msg):
        self._call(
            ["git", "commit", "-m", msg],
            "Could not commit at %(path)s")

    def revert(self):
        self._call(
            ["git", "reset", "--hard"],
            "Could not revert at %(path)s")

    def tag(self, name):
        self._call(
            ["git", "tag", name],
            "Could not tag at %(path)s")

    branch = update = pull = None


class GitCharmTest(Base):

    def setUp(self):
        # Isolate tests from the user's git config.
        orig_home = os.environ.get('HOME')
        os.environ['HOME'] = self.mkdir()
        self.addCleanup(os.environ.update, {'HOME': orig_home})

        self.repo_path = d = self.mkdir()
        self.series_path = os.path.join(d, "precise")
        os.mkdir(self.series_path)
        self.output = self.capture_logging(
            "deployer.charm", level=logging.DEBUG)

    def setup_vcs_charm(self):
        self.branch = Git(self.mkdir())
        self.branch.init()
        self.branch.write(
            {'metadata.yaml': yaml_dump({
                'name': 'couchdb',
                'summary': 'RESTful document oriented database',
                'provides': {
                    'db': {
                        'interface': 'couchdb'}}}),
             'revision': '3'})
        self.branch.commit('initial')
        self.branch.write({'revision': '4'})
        self.branch.commit('next')
        self.branch.tag('v2')
        self.tagged_revision = self.branch.get_cur_rev()
        self.branch.write({'revision': '5'})
        self.branch.commit('next')

        self.charm_data = {
            "charm": "couchdb",
            "build": None,
            "branch": self.branch.path,
            "rev": None,
            "charm_url": None,
        }

    def test_vcs_charm(self):
        self.setup_vcs_charm()
        params = dict(self.charm_data)
        charm = Charm.from_service(
            "scratch", self.repo_path, "precise", params)
        charm.fetch()
        self.assertEqual(charm.metadata['name'], 'couchdb')
        HEAD = charm.vcs.get_cur_rev()

        self.assertFalse(charm.is_modified())
        with open(os.path.join(charm.path, 'revision'), 'w') as fh:
            fh.write('0')
        self.assertTrue(charm.is_modified())
        Git(charm.path).revert()

        charm.rev = None
        # Update goes to latest with no revision specified
        charm.update()
        self.assertEqual(charm.vcs.get_cur_rev(), HEAD)

    def test_vcs_fetch_with_rev(self):
        self.setup_vcs_charm()
        params = dict(self.charm_data)
        rev2 = self.branch._call(
            "git rev-parse HEAD~1".split(),
            self.branch.err_cur_rev,
        )
        params['branch'] = '{}@{}'.format(params['branch'], rev2)
        charm = Charm.from_service(
            "scratch", self.repo_path, "precise", params)
        charm.fetch()
        self.assertEqual(charm.vcs.get_cur_rev(), rev2)

    def test_vcs_fetch_with_tag(self):
        self.setup_vcs_charm()
        params = dict(self.charm_data)
        params['branch'] = '{}@{}'.format(params['branch'], 'v2')
        charm = Charm.from_service(
            "scratch", self.repo_path, "precise", params)
        charm.fetch()
        self.assertEqual(charm.vcs.get_cur_rev(), self.tagged_revision)

    def test_vcs_with_extended_options(self):
        self.setup_vcs_charm()
        params = dict(self.charm_data)
        params['branch'] = \
            "https://review.openstack.org/openstack/charm-ceilometer" + \
            "#changeref=286669/1"

        charm = Charm.from_service(
            "scratch", self.repo_path, "trusty", params)

        self.assertEqual(charm.vcs.extended_options,
                         {'changeref': "286669/1"})

    def test_charm_vcs_unknown(self):
        branch = self.mkdir()
        params = {
            'charm': 'couchdb',
            'branch': "%s" % branch}
        try:
            Charm.from_service(
                "scratch", self.repo_path, "precise", params)
            self.fail("should have failed, vcs ambigious")
        except ValueError as e:
            self.assertIn("Could not determine vcs backend", str(e))



from __future__ import absolute_import
import logging

from ..utils import (
    AlternateKeyDict,
    ErrorExit,
    DeploymentError,
    yaml_load,
    yaml_dump,
    temp_file,
    _check_call,
    get_juju_major_version,
)


class BaseEnvironment(object):

    log = logging.getLogger("deployer.env")

    def __init__(self):
        self.juju_version = get_juju_major_version()

    def _check_call(self, *args, **kwargs):
        if self.options and self.options.retry_count:
            kwargs['max_retry'] = self.options.retry_count
        return _check_call(*args, **kwargs)

    def _named_env(self, params):
        if self.juju_version == 1:
            flag = "-e"
        else:
            flag = "-m"
        if self.name:
            params.extend([flag, self.name])
        return params

    def _write_config(self, svc_name, config, fh):
        fh.write(yaml_dump({svc_name: config}))
        fh.flush()

    def _get_agent_state(self, entity):
        try:
            return entity['agent-state']
        except KeyError:
            # 'agent-state' has been removed for new versions of Juju. Respond
            # with the closest status parameter.
            return entity['agent-status']['status']

    def _get_units_in_error(self, status=None):
        units = []
        if status is None:
            status = self.status()
        for s in status.get('services', {}).keys():
            for uid, u in (status['services'][s].get('units') or {}).items():
                if 'error' in self._get_agent_state(u):
                    units.append(uid)
                for uid, u in (u.get('subordinates') or {}).items():
                    if 'error' in self._get_agent_state(u):
                        units.append(uid)
        return units

    def bootstrap(self, constraints=None):
        self.log.info("bootstrapping, this might take a while...")
        params = ["juju", "bootstrap"]
        if constraints:
            params.extend(['--constraints', constraints])
        self._named_env(params)
        self._check_call(
            params, self.log, "Failed to bootstrap")
        # block until topology is returned
        self.get_cli_status()
        self.log.info(" Bootstrap complete")

    def deploy(self, name, charm_url, repo=None, config=None, constraints=None,
               resources=None, storage=None, num_units=1, force_machine=None,
               series=None, bindings=None):
        params = self._named_env(["juju", "deploy"])
        with temp_file() as fh:
            if config:
                fh.write(yaml_dump({name: config}).encode())
                fh.flush()
                params.extend(["--config", fh.name])
            if constraints:
                if isinstance(constraints, list):
                    constraints = ' '.join(constraints)
                if isinstance(constraints, dict):
                    constraints = ' '.join([
                        '{}={}'.format(k, v) for k, v in constraints.items()
                    ])
                params.extend(['--constraints', constraints])
            if resources:
                if not isinstance(resources, dict):
                    raise DeploymentError(
                        "resources must be specified as a dictionary")
                for key, value in resources.items():
                    params.append("--resource")
                    params.append("{}={}".format(key, value))
            if storage:
                if not isinstance(storage, dict):
                    raise DeploymentError(
                        "storage must be specified as a dictionary")
                params.append("--storage")
                for key, value in storage.items():
                    params.append("{}={}".format(key, value))
            if num_units not in (1, None):
                params.extend(["--num-units", str(num_units)])
            if charm_url.startswith('local'):
                if repo == "":
                    repo = "."
                params.extend(["--repository=%s" % repo])
            if force_machine is not None:
                params.extend(["--to=%s" % force_machine])
            if series and get_juju_major_version() > 1:
                params.extend(['--series', series])
            if bindings:
                if not isinstance(bindings, dict):
                    raise DeploymentError(
                        "bindings must be specified as a dictionary")
                bindings_list = []
                for key, value in bindings.items():
                    if key == "":
                        bindings_list.append("{}".format(value))
                    else:
                        bindings_list.append("{}={}".format(key, value))
                params.extend(["--bind", "%s" % " ".join(bindings_list)])

            params.extend([charm_url, name])
            self._check_call(
                params, self.log, "Error deploying application %r", name)

    def expose(self, name):
        params = self._named_env(["juju", "expose", name])
        self._check_call(
            params, self.log, "Error exposing application %r", name)

    def terminate_machine(self, mid, wait=False, force=False):
        """Terminate a machine.

        Unless ``force=True``, the machine can't have any running units.
        After removing the units or destroying the application, use
        wait_for_units to know when its safe to delete the machine (i.e.,
        units have finished executing stop hooks and are removed).

        """
        if (self.juju_version == 1 and
            ((isinstance(mid, int) and mid == 0) or
             (mid.isdigit() and int(mid) == 0))):
            # Don't kill state server
            raise RuntimeError("Can't terminate machine 0")
        if get_juju_major_version() == 1:
            terminate_action = "terminate-machine"
        else:
            terminate_action = "remove-machine"
        params = self._named_env(["juju", terminate_action])
        params.append(mid)
        if force:
            params.append('--force')
        try:
            self._check_call(
                params, self.log, "Error terminating machine %r" % mid)
        except ErrorExit as e:
            if ("machine %s does not exist" % mid) in e.error.output:
                return
            raise

    def get_service_address(self, svc_name):
        status = self.get_cli_status()
        if svc_name not in status['services']:
            self.log.warning("Application %s does not exist", svc_name)
            return None
        svc = status['services'][svc_name]
        if 'subordinate-to' in svc:
            ps = svc['subordinate-to'][0]
            self.log.info(
                'Application %s is a subordinate to %s, '
                'finding principle application'
                % (svc_name, ps))
            return self.get_service_address(svc['subordinate-to'][0])

        units = svc.get('units') or {}
        unit_keys = list(sorted(units.keys()))
        if unit_keys:
            return units[unit_keys[0]].get('public-address', '')
        self.log.warning("Application %s has no units" % svc_name)

    def get_cli_status(self):
        params = self._named_env(["juju", "status", "--format=yaml"])
        with open('/dev/null', 'w') as fh:
            output = self._check_call(
                params, self.log, "Error getting status, is it bootstrapped?",
                stderr=fh)
        status = yaml_load(output)
        return NormalizedStatus(status)

    def add_unit(self, service_name, machine_spec):
        raise NotImplementedError()

    def set_annotation(self, entity, annotations, entity_type='service'):
        raise NotImplementedError()


class NormalizedStatus(AlternateKeyDict):
    alternates = {
        'services': ('Services', 'applications'),
        'machines': ('Machines',),
        'units': ('Units',),
        'relations': ('Relations',),
        'subordinates': ('Subordinates',),
        'agent-state': ('AgentState',),
    }



"""GUI server environment implementation.

The environment defined here is intended to be used by the Juju GUI server.
See <https://code.launchpad.net/~juju-gui/charms/precise/juju-gui/trunk>.
"""

from .go import GoEnvironment
from ..utils import get_qualified_charm_url, parse_constraints


class GUIEnvironment(GoEnvironment):
    """A Juju environment for the juju-deployer.

    Add support for deployments via the Juju API and for authenticating with
    the provided credentials.
    """

    def __init__(self, endpoint, username, password):
        super(GUIEnvironment, self).__init__('gui', endpoint=endpoint)
        self._username = username
        self._password = password

    def connect(self):
        """Connect the API client to the Juju backend.

        This method is overridden so that a call to connect is a no-op if the
        client is already connected.
        """
        if self.client is None:
            self.client = self.client_class(self.api_endpoint)
            self.client.login(self._password, user=self._username)

    def close(self):
        """Close the API connection.

        Also set the client attribute to None after the disconnection.
        """
        super(GUIEnvironment, self).close()
        self.client = None

    def deploy(
            self, name, charm_url, repo=None, config=None, constraints=None,
            num_units=1, force_machine=None, series=None):
        """Deploy a service using the API.

        Using the API in place of the command line introduces some limitations:
          - it is not possible to use a local charm/repository.

        The repo and series arguments are ignored but listed since the
        Importer always passes the value as a positional argument.

        """
        charm_url = get_qualified_charm_url(charm_url)
        constraints = parse_constraints(constraints)
        self.client.deploy(
            name, charm_url, config=config, constraints=constraints,
            num_units=num_units, machine_spec=force_machine)



from __future__ import absolute_import
from __future__ import print_function
import logging
import time

from .base import BaseAction
from ..relation import EndpointPair
from ..utils import parse_constraints, yaml_dump
import six


class Diff(BaseAction):

    log = logging.getLogger("deployer.diff")

    def __init__(self, env, deployment, options):
        self.options = options
        self.env = env
        self.deployment = deployment
        self.env_status = None
        self.env_state = {'services': {}, 'relations': []}

    def load_env(self):
        """
        """
        rels = set()
        for svc_name in self.env_status['services']:
            if svc_name not in self.env_status['services']:
                self.env_state['services'][svc_name] = 'missing'
            self.env_state['services'].setdefault(svc_name, {})[
                'options'] = self.env.get_config(svc_name)
            self.env_state['services'][svc_name][
                'constraints'] = self.env.get_constraints(svc_name)
            self.env_state['services'][svc_name]['unit_count'] = len(
                (self.env_status['services'][svc_name].get('units') or {}))
            rels.update(self._load_rels(svc_name))
        self.env_state['relations'] = sorted(rels)

    def _load_rels(self, svc_name):
        rels = set()
        svc_rels = self.env_status['services'][svc_name].get(
            'relations', {})
        # There is ambiguity here for multiple rels between two
        # services without the relation id, which we need support
        # from core for.
        for r_name, r_svcs in svc_rels.items():
            for r_svc in r_svcs:
                # Skip peer relations
                if r_svc == svc_name:
                    continue
                rr_name = self._get_rel_name(svc_name, r_svc)
                rels.add(
                    tuple(sorted([
                        "%s:%s" % (svc_name, r_name),
                        "%s:%s" % (r_svc, rr_name)])))
        return rels

    def _get_rel_name(self, src, tgt):
        svc_rels = self.env_status['services'][tgt]['relations']
        found = None
        for r, eps in svc_rels.items():
            if src in eps:
                if found:
                    self.deployment.log.warning(
                        "Application: %s will show ambiguous relations to %s"
                        % (src, tgt))
                    # use "*" to note the ambiguity - will be specially
                    # processed by Endpoint comparisons as a kind of
                    # ~wildcard
                    found = "*"
                else:
                    found = r
        return found

    def get_delta(self):
        delta = {}
        rels_delta = self._get_relations_delta()
        if rels_delta:
            delta['relations'] = rels_delta
        svc_delta = self._get_services_delta()
        if svc_delta:
            delta['services'] = svc_delta
        return delta

    def _get_relations_delta(self):
        # Simple endpoint diff, no qualified endpoint checking.

        # Env relations are always qualified (at least in go).
        delta = {}
        env_rels = set(
            EndpointPair(*x) for x in self.env_state.get('relations', ()))
        dep_rels = set(
            [EndpointPair(*y) for y in self.deployment.get_relations()])

        for r in dep_rels.difference(env_rels):
            delta.setdefault('missing', []).append(r)

        for r in env_rels.difference(dep_rels):
            delta.setdefault('unknown', []).append(r)

        return delta

    def _get_services_delta(self):
        delta = {}
        env_svcs = set(self.env_status['services'].keys())
        dep_svcs = set([s.name for s in self.deployment.get_services()])

        missing = dep_svcs - env_svcs
        if missing:
            delta['missing'] = {}
        for a in missing:
            delta['missing'][a] = self.deployment.get_service(
                a).svc_data
        unknown = env_svcs - dep_svcs
        if unknown:
            delta['unknown'] = {}
        for r in unknown:
            delta['unknown'][r] = self.env_state.get(r)

        for cs in env_svcs.intersection(dep_svcs):
            d_s = self.deployment.get_service(cs).svc_data
            e_s = self.env_state['services'][cs]
            mod = self._diff_service(e_s, d_s,
                                     self.deployment.get_charm_for(cs))
            if not mod:
                continue
            if 'modified' not in delta:
                delta['modified'] = {}
            delta['modified'][cs] = mod
        return delta

    def _diff_service(self, env_service, dep_service, charm):
        mod = {}

        # Start by diffing the constraints.
        dep_constraints = parse_constraints(dep_service.get('constraints', ''))
        # 'tags' is a special case, as it can be multi-valued: convert to list
        # if the deployment specifies just a string
        if isinstance(dep_constraints.get('tags'), six.string_types):
            dep_constraints['tags'] = [dep_constraints['tags']]
        if dep_constraints != env_service['constraints']:
            mod['env-constraints'] = env_service['constraints']
            mod['cfg-constraints'] = dep_constraints

        # Now collect all the options from both services and diff them.
        all_options = (set(dep_service.get('options', {}).keys()) |
                       set(env_service.get('options', {}).keys()))
        for key in all_options:
            dep_value = dep_service.get('options', {}).get(key)
            # Assume the charm default if the deployment specifies no value.
            if dep_value is None:
                dep_value = charm.config.get(key, {}).get('default', None)
            env_value = env_service['options'].get(key, {}).get('value')
            if env_value != dep_value:
                mod.setdefault('env-config', {}).update({key: env_value})
                mod.setdefault('cfg-config', {}).update({key: dep_value})

        # Finally, compare the unit counts if appropriate.
        if not charm or not charm.is_subordinate():
            # XXX: The num_units default policy should really be
            # implemented centrally.
            dep_unit_count = dep_service.get('num_units', 1)
            if env_service['unit_count'] != dep_unit_count:
                mod['num_units'] = env_service['unit_count'] - dep_unit_count
        return mod

    def do_diff(self):
        self.start_time = time.time()
        self.deployment.resolve()
        self.env.connect()
        self.env_status = self.env.status()
        self.load_env()
        return self.get_delta()

    def run(self):
        diff = self.do_diff()
        if diff:
            print(yaml_dump(diff))



from __future__ import absolute_import
import os

import mock

from deployer.config import ConfigStack
from deployer.action.importer import Importer

from .base import (
    Base,
    skip_if_offline,
)


class Options(dict):

    def __getattr__(self, key):
        return self[key]


class ImporterTest(Base):

    def setUp(self):
        self.juju_home = self.mkdir()
        self.change_environment(JUJU_HOME=self.juju_home)
        self.options = Options({
            'bootstrap': False,
            'branch_only': False,
            'configs': [os.path.join(self.test_data_dir, 'wiki.yaml')],
            'debug': True,
            'deploy_delay': 0,
            'destroy_services': None,
            'diff': False,
            'find_service': None,
            'ignore_errors': False,
            'list_deploys': False,
            'no_local_mods': True,
            'no_relations': False,
            'overrides': None,
            'rel_wait': 60,
            'retry_count': 0,
            'series': None,
            'skip_unit_wait': False,
            'terminate_machines': False,
            'timeout': 2700,
            'update_charms': False,
            'verbose': True,
            'watch': False})

    @skip_if_offline
    @mock.patch('deployer.action.importer.time')
    def test_importer(self, mock_time):
        mock_time.time.return_value = 0
        # Trying to track down where this comes from http://pad.lv/1243827
        stack = ConfigStack(self.options.configs)
        deploy = stack.get('wiki')
        env = mock.MagicMock()
        importer = Importer(env, deploy, self.options)
        importer.run()

        config = {'name': '$hi_world _are_you_there? {guess_who}'}
        self.assertEqual(
            env.method_calls[3], mock.call.deploy(
                'wiki', 'cs:precise/mediawiki',
                os.environ.get("JUJU_REPOSITORY", ""),
                config, None, None, None, 1, None, 'precise', None))
        env.add_relation.assert_called_once_with('wiki', 'db')

    @skip_if_offline
    @mock.patch('deployer.action.importer.time')
    def test_importer_with_resources(self, mock_time):
        mock_time.time.return_value = 0
        stack = ConfigStack([
            os.path.join(self.test_data_dir, 'wiki-resources.yaml')])
        deploy = stack.get('wiki')
        env = mock.MagicMock()
        importer = Importer(env, deploy, self.options)
        importer.run()

        config = {'name': '$hi_world _are_you_there? {guess_who}'}
        self.assertEqual(
            env.method_calls[3], mock.call.deploy(
                'wiki', 'cs:precise/mediawiki',
                os.environ.get("JUJU_REPOSITORY", ""),
                config, None,
                {"config": "./docs/cfg.xml", "tarball": "/some/file.tgz"},
                None, 1, None, 'precise', None))
        env.add_relation.assert_called_once_with('wiki', 'db')

    @skip_if_offline
    @skip_if_offline
    @mock.patch('deployer.action.importer.time')
    def test_importer_with_storage(self, mock_time):
        mock_time.time.return_value = 0
        stack = ConfigStack([
            os.path.join(self.test_data_dir, 'wiki-storage.yaml')])
        deploy = stack.get('wiki')
        env = mock.MagicMock()
        importer = Importer(env, deploy, self.options)
        importer.run()

        config = {'name': '$hi_world _are_you_there? {guess_who}'}
        self.assertEqual(
            env.method_calls[3], mock.call.deploy(
                'wiki', 'cs:precise/mediawiki',
                os.environ.get("JUJU_REPOSITORY", ""),
                config, None, None, {"web-data": "cinder,10G,1"},
                1, None, 'precise', None))
        env.add_relation.assert_called_once_with('wiki', 'db')

    @skip_if_offline
    @skip_if_offline
    @mock.patch('deployer.action.importer.time')
    def test_importer_with_bindings(self, mock_time):
        mock_time.time.return_value = 0
        stack = ConfigStack([
            os.path.join(self.test_data_dir, 'wiki-bindings.yaml')])
        deploy = stack.get('wiki')
        env = mock.MagicMock()
        importer = Importer(env, deploy, self.options)
        importer.run()

        config = {'name': '$hi_world _are_you_there? {guess_who}'}
        self.assertEqual(
            env.method_calls[3], mock.call.deploy(
                'wiki', 'cs:precise/mediawiki',
                os.environ.get("JUJU_REPOSITORY", ""),
                config, None, None, None,
                1, None, 'precise', {"": "default"}))
        env.add_relation.assert_called_once_with('wiki', 'db')

    @skip_if_offline
    @mock.patch('deployer.action.importer.time')
    def test_importer_no_relations(self, mock_time):
        mock_time.time.return_value = 0
        self.options.no_relations = True
        stack = ConfigStack(self.options.configs)
        deploy = stack.get('wiki')
        env = mock.MagicMock()
        importer = Importer(env, deploy, self.options)
        importer.run()
        self.assertFalse(env.add_relation.called)

    @skip_if_offline
    @mock.patch('deployer.action.importer.time')
    def test_importer_add_machine_series(self, mock_time):
        mock_time.time.return_value = 0
        self.options.configs = [
            os.path.join(self.test_data_dir, 'v4', 'series.yaml')]
        stack = ConfigStack(self.options.configs)
        deploy = stack.get(self.options.configs[0])
        env = mock.MagicMock()
        importer = Importer(env, deploy, self.options)
        importer.run()

        self.assertEqual(env.add_machine.call_count, 2)
        env.add_machine.assert_has_calls([
            mock.call(series='precise', constraints='mem=512M'),
            mock.call(series='trusty', constraints='mem=512M'),
        ], any_order=True)

    @skip_if_offline
    @mock.patch('deployer.action.importer.time')
    def test_importer_existing_machine(self, mock_time):
        mock_time.time.return_value = 0
        self.options.configs = [
            os.path.join(self.test_data_dir, 'v4',
                         'container-existing-machine.yaml')]
        stack = ConfigStack(self.options.configs)
        deploy = stack.get(self.options.configs[0])
        env = mock.MagicMock()

        config = {'status.return_value.get.return_value': {
            1: {}
        }}

        env.configure_mock(**config)

        importer = Importer(env, deploy, self.options)
        importer.run()
        self.assertFalse(env.add_machine.called)



from __future__ import absolute_import
import logging
import os
import time
import sys
import unittest

from deployer.env.go import GoEnvironment

from .base import Base


# Takes roughly about 6m on core2 + ssd, mostly cloudinit time
@unittest.skipIf(
    (not bool(os.environ.get("TEST_ENDPOINT"))),
    "Test env must be defined: TEST_ENDPOINT")
class LiveEnvironmentTest(Base):

    @classmethod
    def setUpClass(cls):
        """Base class sets JUJU_HOME to a new tmp dir, but for these
        tests we need a real JUJU_HOME so that calls to
        ``juju api-endpoints`` work properly.

        """
        if not os.environ.get('JUJU_HOME'):
            raise RuntimeError('JUJU_HOME must be set')

    @classmethod
    def tearDownClass(cls):
        """Base class deletes the tmp JUJU_HOME dir it created.
        Override so we don't delete our real JUJU_HOME!

        """
        pass

    def setUp(self):
        self.endpoint = os.environ.get("TEST_ENDPOINT")
        self.output = self.capture_logging(
            "deployer", log_file=sys.stderr, level=logging.DEBUG)
        self.env = GoEnvironment(
            os.environ.get("JUJU_ENV"), endpoint=self.endpoint)
        self.env.connect()
        status = self.env.status()
        self.assertFalse(status.get('services'))
        # Destroy everything.. consistent baseline
        self.env.reset(
            terminate_machines=len(list(status['machines'].keys())) > 1,
            terminate_delay=240)

    def tearDown(self):
        self.env.reset(
            terminate_machines=True, terminate_delay=240,
            force_terminate=True)
        self.env.close()

    def test_env(self):
        status = self.env.status()
        self.env.deploy("test-blog", "cs:precise/wordpress")
        self.env.deploy("test-db", "cs:precise/mysql")
        self.env.add_relation("test-db", "test-blog")
        self.env.deploy(
            "test-db2", "cs:postgresql",
            storage={"pgdata": "rootfs,10M"}, series="xenial")
        self.env.add_units('test-blog', 1)

        # Sleep because juju core watches are eventually consistent (5s window)
        # and status rpc is broken (http://pad.lv/1203105)
        time.sleep(6)
        self.env.wait_for_units(timeout=800)
        status = self.env.status()

        services = ["test-blog", "test-db", "test-db2"]
        self.assertEqual(
            sorted(status['services'].keys()),
            services)
        for s in services:
            for k, u in status['services'][s]['units'].items():
                try:
                    self.assertIn(u['agent-state'],
                                  ("allocating", "pending", "started"))
                except KeyError:
                    self.assertIn(u['agent-status']['status'],
                                  ("allocating", "idle"))

    def test_add_machine(self):
        machine_name = self.env.add_machine()

        # Sleep because juju core watches are eventually consistent (5s window)
        # and status rpc is broken (http://pad.lv/1203105)
        time.sleep(6)
        status = self.env.status()

        self.assertIn(machine_name, status['machines'])

    def test_set_annotation(self):
        machine_name = self.env.add_machine()
        self.env.set_annotation(
            machine_name, {'foo': 'bar'}, entity_type='machine')

        # Sleep because juju core watches are eventually consistent (5s window)
        # and status rpc is broken (http://pad.lv/1203105)
        time.sleep(6)
        self.env.status()

        self.assertIn('foo', self.env.client.get_annotation(
            machine_name, 'machine')['Annotations'])



from __future__ import absolute_import
from deployer.service import Service
from .base import Base


class ServiceTest(Base):

    def test_service(self):
        data = {
            'branch': 'lp:precise/mysql'}

        s = Service('db', data)
        self.assertEqual(s.name, "db")
        self.assertEqual(s.num_units, 1)
        self.assertEqual(s.constraints, None)
        self.assertEqual(s.config, None)
        self.assertEqual(s.storage, None)

        data = {
            'branch': 'lp:precise/mysql',
            'constraints': "instance-type=m1.small",
            'options': {"services": "include-file://stack-include.yaml"},
            'num_units': 10,
            'resources': {"config": "./docs/cfg.xml"},
            'storage': {"data": "ebs,10G,1"}}
        s = Service('db', data)
        self.assertEquals(s.num_units, 10)
        self.assertEquals(s.constraints, "instance-type=m1.small")
        self.assertEquals(
            s.config,
            {"services": "include-file://stack-include.yaml"})
        self.assertEquals(s.resources, {"config": "./docs/cfg.xml"})
        self.assertEquals(s.storage, {"data": "ebs,10G,1"})



from __future__ import absolute_import
import logging
import mock
import os
import tempfile
import yaml

import six

from deployer.deployment import Deployment
from deployer.config import ConfigStack
from deployer.utils import ErrorExit

from .base import Base


class ConfigTest(Base):

    def setUp(self):
        self.output = self.capture_logging(
            "deployer.config", level=logging.DEBUG)

    def test_config_basic(self):
        config = ConfigStack(['configs/ostack-testing-sample.cfg'])
        config.load()
        self.assertEqual(
            list(config.keys()),
            [u'openstack-precise-ec2',
             u'openstack-precise-ec2-trunk',
             u'openstack-ubuntu-testing'])
        self.assertRaises(ErrorExit, config.get, 'zeeland')
        result = config.get("openstack-precise-ec2")
        self.assertTrue(isinstance(result, Deployment))
        self.assertEqual(config.version, 3)

    def test_config(self):
        config = ConfigStack([
            os.path.join(self.test_data_dir, "stack-default.cfg"),
            os.path.join(self.test_data_dir, "stack-inherits.cfg")])
        config.load()
        self.assertEqual(
            list(config.keys()),
            [u'my-files-frontend-dev', u'wordpress'])
        deployment = config.get("wordpress")
        self.assertTrue(deployment)
        self.assertEqual(config.version, 3)

    def test_config_v4(self):
        config = ConfigStack([
            os.path.join(self.test_data_dir, 'v4', 'simple.yaml')])
        config.load()
        self.assertEqual(
            list(config.keys()),
            [os.path.join(self.test_data_dir, 'v4', 'simple.yaml')])
        with mock.patch('deployer.config.ConfigStack._resolve_inherited') \
                as mock_resolve:
            deployment = config.get(list(config.keys())[0])
            self.assertTrue(deployment)
            self.assertFalse(mock_resolve.called)
        self.assertEqual(config.version, 4)

    def test_config_include_file(self):
        config = ConfigStack([
            os.path.join(self.test_data_dir, "stack-includes.cfg")])
        config.load()
        # ensure picked up stacks from both files
        self.assertEqual(
            list(config.keys()),
            [u'my-files-frontend-dev', u'wordpress'])

        # ensure inheritance was adhered to during cross-file load
        wordpress = [s.name for s in config.get('wordpress').get_services()]
        my_app = [s.name for s in
                  config.get('my-files-frontend-dev').get_services()]
        self.assertTrue(set(wordpress).issubset(set(my_app)))

    def test_inherits_config_overridden(self):
        config = ConfigStack([
            os.path.join(self.test_data_dir, "stack-default.cfg"),
            os.path.join(self.test_data_dir, "stack-inherits.cfg")])
        config.load()
        deployment = config.get('my-files-frontend-dev')
        db = deployment.get_service('db')
        # base deployment (wordpress)'s db tuning level should have been
        # over-ridden
        self.assertEquals(db.config.get('tuning-level'), 'fastest')

    def test_multi_inheritance_multi_files(self):
        config = ConfigStack([
            os.path.join(self.test_data_dir, "openstack", "openstack.cfg"),
            os.path.join(self.test_data_dir, "openstack", "ubuntu_base.cfg"),
            os.path.join(
                self.test_data_dir, "openstack", "openstack_base.cfg"),
        ])
        self._test_multiple_inheritance(config)

    def test_multi_inheritance_multi_included_files(self):
        # openstack.cfg
        #  includes -> ubuntu_base.cfg includes
        #               includes -> openstack_base.cfg
        test_conf = yaml.load(open(
            os.path.join(self.test_data_dir, "openstack", "openstack.cfg")))
        includes = [
            os.path.join(self.test_data_dir, "openstack", "ubuntu_base.cfg"),
            os.path.join(self.test_data_dir, "openstack", "openstack_base.cfg")
        ]
        for key in ['include-config', 'include-configs']:
            test_conf[key] = includes
            with tempfile.NamedTemporaryFile() as tmp_cfg:
                tmp_cfg.write(yaml.dump(test_conf).encode())
                tmp_cfg.flush()
                config = ConfigStack([tmp_cfg.name])
                self._test_multiple_inheritance(config)
            del test_conf[key]

    def test_multi_inheritance_included_multi_configs(self):
        # openstack.cfg
        #  includes -> [ubuntu_base.cfg, openstack_base.cfg]
        config = ConfigStack([
            os.path.join(self.test_data_dir, "openstack", "openstack.cfg"),
        ])
        self._test_multiple_inheritance(config)

    def _test_multiple_inheritance(self, config):
        config.load()

        deployment = config.get('precise-grizzly')
        services = [s.name for s in list(deployment.get_services())]
        self.assertEquals(['mysql', 'nova-cloud-controller'], services)

        nova = deployment.get_service('nova-cloud-controller')
        self.assertEquals(nova.config['openstack-origin'],
                          'cloud:precise-grizzly')

        deployment = config.get('precise-grizzly-quantum')
        services = [s.name for s in list(deployment.get_services())]
        self.assertEquals(services,
                          ['mysql', 'nova-cloud-controller',
                           'quantum-gateway'])
        nova = deployment.get_service('nova-cloud-controller')
        self.assertEquals(nova.config['network-manager'], 'Quantum')
        self.assertEquals(nova.config['openstack-origin'],
                          'cloud:precise-grizzly')
        ex_rels = [('quantum-gateway', 'nova-cloud-controller'),
                   ('quantum-gateway', 'mysql'),
                   ('nova-cloud-controller', 'mysql')]
        self.assertEquals(
            sorted(ex_rels),
            sorted(list(deployment.get_relations())))

    def test_config_series_override(self):
        config = ConfigStack(['configs/wiki.yaml'], 'trusty')
        config.load()
        result = config.get("wiki")
        self.assertTrue(isinstance(result, Deployment))
        self.assertEquals(result.series, 'trusty')


class NetworkConfigFetchingTests(Base):
    """Configuration files can be specified via URL that is then fetched."""

    def setUp(self):
        self.output = self.capture_logging(
            "deployer.config", level=logging.DEBUG)

    def test_urls_are_fetched(self):
        # If a config file is specified as a URL, that URL is fetched and
        # placed at a temporary location where it is read and treated as a
        # regular config file.
        CONFIG_URL = 'http://site.invalid/config-1'
        config = ConfigStack([])
        config.config_files = [CONFIG_URL]

        class FauxResponse(six.BytesIO):
            def getcode(self):
                return 200

        def faux_urlopen(url):
            self.assertEqual(url, CONFIG_URL)
            with open('configs/ostack-testing-sample.cfg', 'rb') as f:
                return FauxResponse(f.read())

        config.urlopen = faux_urlopen
        config.load()
        self.assertEqual(
            list(config.keys()),
            [u'openstack-precise-ec2',
             u'openstack-precise-ec2-trunk',
             u'openstack-ubuntu-testing'])
        self.assertRaises(ErrorExit, config.get, 'zeeland')
        result = config.get("openstack-precise-ec2")
        self.assertTrue(isinstance(result, Deployment))

    def test_unfetchable_urls_generate_an_error(self):
        # If a config file is specified as a URL, that URL is fetched and
        # placed at a temporary location where it is read and treated as a
        # regular config file.
        CONFIG_URL = 'http://site.invalid/config-1'
        config = ConfigStack([])
        config.config_files = [CONFIG_URL]

        class FauxResponse(six.BytesIO):
            def getcode(self):
                return 400

        def faux_urlopen(url):
            self.assertEqual(url, CONFIG_URL)
            with open('configs/ostack-testing-sample.cfg', 'rb') as f:
                return FauxResponse(f.read())

        config.urlopen = faux_urlopen
        self.assertRaises(ErrorExit, config.load)



from __future__ import absolute_import
import itertools

from .feedback import Feedback
from .utils import get_juju_major_version
from six.moves import map

# Map of container-type-used-in-bundle to actual-container-type-passed-to-juju.
if get_juju_major_version() == 1:
    # lxd not supported on juju1
    CONTAINER_TYPES = {
        'lxc': 'lxc',
        'kvm': 'kvm',
    }
else:
    # If you use 'lxc' in a bundle with juju2, deployer will translate it to
    # 'lxd' before passing the deploy command to juju.
    CONTAINER_TYPES = {
        'lxc': 'lxd',
        'lxd': 'lxd',
        'kvm': 'kvm',
    }


class Service(object):

    def __init__(self, name, svc_data):
        self.svc_data = svc_data
        self.name = name

    def __repr__(self):
        return "<Service %s>" % (self.name)

    @property
    def annotations(self):
        a = self.svc_data.get('annotations')
        if a is None:
            return a
        # core annotations only supports string key / values
        d = {}
        for k, v in a.items():
            d[str(k)] = str(v)
        return d

    @property
    def config(self):
        return self.svc_data.get('options', None)

    @property
    def constraints(self):
        return self.svc_data.get('constraints', None)

    @property
    def num_units(self):
        return int(self.svc_data.get('num_units', 1))

    @property
    def resources(self):
        return self.svc_data.get('resources', None)

    @property
    def bindings(self):
        return self.svc_data.get('bindings', None)

    @property
    def storage(self):
        return self.svc_data.get('storage', None)

    @property
    def unit_placement(self):
        # Separate checks to support machine 0 placement.
        value = self.svc_data.get('to')
        if value is None:
            value = self.svc_data.get('force-machine')
        if value is not None and not isinstance(value, list):
            value = [value]
        return value and list(map(str, value)) or []

    @property
    def expose(self):
        return self.svc_data.get('expose', False)


class ServiceUnitPlacement(object):

    def __init__(self, service, deployment, status, arbitrary_machines=True):
        self.service = service
        self.deployment = deployment
        self.status = status
        self.arbitrary_machines = arbitrary_machines

    @staticmethod
    def _format_placement(machine, container=None):
        if container:
            return "%s:%s" % (container, machine)
        else:
            return machine

    def colocate(self, status, placement, u_idx, container, svc):
        """Colocate one service with an existing service either within a
        container alongside that service or hulk-smashed onto the same unit.

        status: the environment status.
        placement: the placement directive of the unit to be colocated.
        u_idx: the unit index of the unit to be colocated.
        container: a string containing the container type, or None.
        svc: the service object for this placement.
        """
        with_service = status['services'].get(placement)
        if with_service is None:
            # Should be caught in validate relations but sanity check
            # for concurrency.
            self.deployment.log.error(
                "Application %s to be deployed with non-existent "
                "application %s", svc.name, placement)
            # Prefer continuing deployment with a new machine rather
            # than an in-progress abort.
            return None

        svc_units = with_service['units']
        if int(u_idx) >= len(svc_units):
            self.deployment.log.warning(
                "Application:%s, Deploy-with-application:%s, "
                "Requested-unit-index=%s, "
                "Cannot solve, falling back to default placement",
                svc.name, placement, u_idx)
            return None
        unit_names = list(svc_units.keys())
        unit_names.sort(key=lambda unit: int(unit.split('/')[1]))
        machine = (
            svc_units[unit_names[int(u_idx)]].get('Machine') or
            svc_units[unit_names[int(u_idx)]].get('machine')
        )
        if not machine:
            self.deployment.log.warning(
                "Application:%s deploy-with unit missing machine for %s",
                svc.name, unit_names[int(u_idx)])
            return None
        return self._format_placement(machine, container)


class ServiceUnitPlacementV3(ServiceUnitPlacement):

    def _parse_placement(self, unit_placement):
        placement = unit_placement
        container = None
        u_idx = None
        if ':' in unit_placement:
            container, placement = unit_placement.split(":")
        if '=' in placement:
            placement, u_idx = placement.split("=")
        return container, placement, u_idx

    def validate(self):
        feedback = Feedback()

        unit_placement = self.service.unit_placement
        if unit_placement is None:
            return feedback

        if not isinstance(unit_placement, list):
            unit_placement = [unit_placement]
        unit_placement = list(map(str, unit_placement))

        services = dict([(s.name, s) for s in self.deployment.get_services()])
        machines = self.deployment.get_machines()

        for idx, p in enumerate(unit_placement):
            container, p, u_idx = self._parse_placement(p)
            if container and container not in CONTAINER_TYPES:
                feedback.error(
                    "Invalid container type:%s application: %s placement: %s"
                    % (container, self.service.name, unit_placement[idx]))
            if u_idx:
                if p in ('maas', 'zone'):
                    continue
                if not u_idx.isdigit():
                    feedback.error(
                        "Invalid application:%s placement: %s" % (
                            self.service.name, unit_placement[idx]))
            if p.isdigit():
                if p == '0' or p in machines or self.arbitrary_machines:
                    continue
                else:
                    feedback.error(
                        ("Application placement to machine "
                         "not supported %s to %s") % (
                            self.service.name, unit_placement[idx]))
            elif p in services:
                if services[p].unit_placement:
                    feedback.error(
                        "Nested placement not supported %s -> %s -> %s" % (
                            self.service.name, p, services[p].unit_placement))
                elif self.deployment.get_charm_for(p).is_subordinate():
                    feedback.error(
                        "Cannot place to a subordinate application: "
                        "%s -> %s" % (self.service.name, p))
            else:
                feedback.error(
                    "Invalid application placement %s to %s" % (
                        self.service.name, unit_placement[idx]))
        return feedback

    def get(self, unit_number):
        """Get the placement directive for a given unit.

        unit_number: the number of the unit to deploy
        """
        status = self.status
        svc = self.service

        unit_mapping = svc.unit_placement
        if not unit_mapping:
            return None
        if len(unit_mapping) <= unit_number:
            return None

        unit_placement = placement = str(unit_mapping[unit_number])
        container = None
        u_idx = unit_number

        if ':' in unit_placement:
            container, placement = unit_placement.split(":")
            container = CONTAINER_TYPES[container]
        if '=' in placement:
            placement, u_idx = placement.split("=")

        if placement.isdigit():
            if self.arbitrary_machines or placement == '0':
                return self._format_placement(placement, container)
        if placement == 'maas':
            return u_idx
        elif placement == "zone":
            return "zone=%s" % u_idx

        return self.colocate(status, placement, u_idx, container, svc)


class ServiceUnitPlacementV4(ServiceUnitPlacement):

    def __init__(self, service, deployment, status, arbitrary_machines=False,
                 machines_map=None):
        super(ServiceUnitPlacementV4, self).__init__(
            service, deployment, status, arbitrary_machines=arbitrary_machines)

        # Arbitrary machines will not be allowed in v4 bundles.
        self.arbitrary_machines = False

        self.machines_map = machines_map

        # Ensure that placement spec is filled according to the bundle
        # specification.
        self._fill_placement()

    def _fill_placement(self):
        """Fill the placement spec with necessary data.

        From the spec:
        A unit placement may be specified with an application name only,
        in which case its unit number is assumed to be one more than the
        unit number of the previous unit in the list with the same
        application, or zero if there were none.

        If there are less elements in To than NumUnits, the last element is
        replicated to fill it. If there are no elements (or To is omitted),
        "new" is replicated.

        """
        unit_mapping = self.service.unit_placement
        unit_count = self.service.num_units
        if not unit_mapping:
            self.service.svc_data['to'] = ['new'] * unit_count
            return

        self.service.svc_data['to'] = (
            unit_mapping +
            list(itertools.repeat(
                unit_mapping[-1], unit_count - len(unit_mapping)))
        )
        unit_mapping = self.service.unit_placement

        colocate_counts = {}
        for idx, mapping in enumerate(unit_mapping):
            service = mapping
            if ':' in mapping:
                service = mapping.split(':')[1]
            if service in self.deployment.data['services']:
                unit_number = colocate_counts.setdefault(service, 0)
                unit_mapping[idx] = "{}/{}".format(mapping, unit_number)
                colocate_counts[service] += 1
        self.service.svc_data['to'] = unit_mapping

    def _parse_placement(self, placement):
        """Parse a unit placement statement.

        In version 4 bundles, unit placement statements take the form of

          (<containertype>:)?(<unit>|<machine>|new)

        This splits the placement into a container, a placement, and a unit
        number.  Both container and unit number are optional and can be None.

        """
        container = unit_number = None
        if ':' in placement:
            container, placement = placement.split(':')
            container = CONTAINER_TYPES.get(container, container)
        if '/' in placement:
            placement, unit_number = placement.split('/')
        return container, placement, unit_number

    def validate(self):
        """Validate the placement of an application and all of its units.

        If an application has a 'to' block specified, the list of machines,
        units, containers, and/or applications must be internally consistent,
        consistent with other applications in the deployment, and consistent
        with any machines specified in the 'machines' block of the deployment.

        A feedback object is returned, potentially with errors and warnings
        inside it.

        """
        feedback = Feedback()

        unit_placement = self.service.unit_placement
        if unit_placement is None:
            return feedback

        if not isinstance(unit_placement, (list, tuple)):
            unit_placement = [unit_placement]
        unit_placement = list(map(str, unit_placement))

        services = dict([(s.name, s) for s in self.deployment.get_services()])
        machines = self.deployment.get_machines()
        service_name = self.service.name

        for i, placement in enumerate(unit_placement):
            container, target, unit_number = self._parse_placement(placement)

            # Validate the container type.
            if container and container not in CONTAINER_TYPES:
                feedback.error(
                    'Invalid container type: {} application: {} placement: {}'
                    ''.format(container, service_name, placement))
            # Specify an existing machine (or, if the number is in the
            # list of machine specs, one of those).
            if str(target) in machines:
                continue
            if target.isdigit():
                feedback.error(
                    'Application placement to machine not supported: {} to {}'
                    ''.format(service_name, placement))
            # Specify a service for co-location.
            elif target in services:
                # Specify a particular unit for co-location.
                if unit_number is not None:
                    try:
                        unit_number = int(unit_number)
                    except (TypeError, ValueError):
                        feedback.error(
                            'Invalid unit number for placement: {} to {}'
                            ''.format(service_name, unit_number))
                        continue
                    if unit_number > services[target].num_units:
                        feedback.error(
                            'Application unit does not exist: {} to {}/{}'
                            ''.format(service_name, target, unit_number))
                        continue
                if self.deployment.get_charm_for(target).is_subordinate():
                    feedback.error(
                        'Cannot place to a subordinate application: {} -> {}'
                        ''.format(service_name, target))
            # Create a new machine or container.
            elif target == 'new':
                continue
            else:
                feedback.error(
                    'Invalid application placement: {} to {}'
                    ''.format(service_name, placement))
        return feedback

    def get_new_machines_for_containers(self):
        """Return a list of containers in the application's unit placement that
        have been requested to be put on new machines."""
        new_machines = []
        unit = itertools.count()
        for placement in self.service.unit_placement:
            if ':new' in placement:
                # Generate a name for this machine to be used in the
                # machines_map used later; as a quick path forward, simply use
                # the unit's name.
                new_machines.append('{}/{}'.format(
                    self.service.name, next(unit)))
        return new_machines

    def get(self, unit_number):
        """Get the placement directive for a given unit.

        unit_number: the number of the unit to deploy
        """
        status = self.status
        svc = self.service

        unit_mapping = svc.unit_placement
        if not unit_mapping:
            return None
        unit_placement = placement = str(unit_mapping[unit_number])
        container = None
        u_idx = unit_number

        # Shortcut for new machines.
        if placement == 'new':
            return None

        container, placement, unit_number = self._parse_placement(
            unit_placement)

        if placement in self.machines_map:
            return self._format_placement(
                self.machines_map[placement], container)

        # Handle <container_type>:new
        if placement == 'new':
            return self._format_placement(
                self.machines_map['%s/%d' % (self.service.name, u_idx)],
                container)

        return self.colocate(status, placement, u_idx, container, svc)



from __future__ import absolute_import
import yaml


class Endpoint(object):

    def __init__(self, ep):
        self.ep = ep
        self.name = None
        if ":" in self.ep:
            self.service, self.name = self.ep.split(":")
        else:
            self.service = ep

    def __eq__(self, ep):
        # exactly equal by full endpoint (service:name), or
        # equal only by service if name is "*" or unset
        return (
            (self.ep == ep.ep) or (
                (not self.name or self.name == "*" or
                    not ep.name or ep.name == "*") and
                self.service == ep.service
            )
        )


class EndpointPair(object):
    # Really simple endpoint service matching that does not work for multiple
    # relations between two services (used by diff at the moment)

    def __init__(self, ep_x, ep_y=None):
        self.ep_x = Endpoint(ep_x)
        self.ep_y = ep_y and Endpoint(ep_y)

    def __eq__(self, ep_pair):
        if not isinstance(ep_pair, EndpointPair):
            return False
        return (ep_pair.ep_x in self and
                ep_pair.ep_y in self)

    def __contains__(self, ep):
        return (self.ep_x == ep or self.ep_y == ep)

    def __hash__(self):
        return hash(tuple(sorted(
            (self.ep_x.service, self.ep_y.service))))

    def __repr__(self):
        return "%s <-> %s" % (
            self.ep_x.ep,
            self.ep_y.ep)

    @staticmethod
    def to_yaml(dumper, data):
        return dumper.represent_list([[data.ep_x.ep, data.ep_y.ep]])


yaml.add_representer(EndpointPair, EndpointPair.to_yaml)



""" Unittest for juju-deployer diff action (--diff) """
# pylint: disable=C0103
from __future__ import absolute_import
import os
import shutil
import tempfile
import unittest
import yaml

from mock import patch, mock_open
from six import StringIO

from deployer.config import ConfigStack
from deployer.env.mem import MemoryEnvironment
from deployer.utils import setup_logging

from .base import Base, skip_if_offline, TEST_OFFLINE, TEST_OFFLINE_REASON
from ..action.diff import Diff


@skip_if_offline
class DiffTest(Base):

    def setUp(self):
        self.output = setup_logging(
            debug=True, verbose=True, stream=StringIO())

    # Because fetch_charms is expensive, do it once for all tests
    @classmethod
    def setUpClass(cls):
        super(DiffTest, cls).setUpClass()
        # setUpClass not being skipped, here, could have to do with
        # decorator on derived class.  So skip explicitly.
        if TEST_OFFLINE:
            raise unittest.SkipTest(TEST_OFFLINE_REASON)
        deployment = ConfigStack(
            [os.path.join(
                cls.test_data_dir,
                "openstack",
                "openstack-1501.yaml")]).get('openstack')
        cls._dir = tempfile.mkdtemp()
        os.mkdir(os.path.join(cls._dir, "trusty"))
        deployment.repo_path = cls._dir
        deployment.fetch_charms()
        deployment.resolve()
        cls._deployment = deployment

    @classmethod
    def tearDownClass(cls):
        super(DiffTest, cls).tearDownClass()
        # TearDownClass not being skipped, here, could have to do with
        # decorator on derived class.  So skip explicitly.
        if TEST_OFFLINE:
            raise unittest.SkipTest(TEST_OFFLINE_REASON)
        shutil.rmtree(cls._dir)

    @classmethod
    def get_dir(cls):
        """ Return temp working dir
        """
        return cls._dir

    @classmethod
    def get_deployment(cls):
        """ Return saved deployment at class initialization
        """
        return cls._deployment

    def test_diff_nil(self):
        dpl = self.get_deployment()
        # No changes, assert nil diff
        env = MemoryEnvironment(dpl.name, dpl)
        diff = Diff(env, dpl, {}).do_diff()
        self.assertEqual(diff, {})

    def test_diff_num_units(self):
        # Removing 1 unit must show -1 'num_units'
        dpl = self.get_deployment()
        env = MemoryEnvironment(dpl.name, dpl)
        env.remove_unit(env.status()['services']['nova-compute']['units'][0])
        diff = Diff(env, dpl, {}).do_diff()
        self.assertEqual(
            diff['services']['modified']['nova-compute']['num_units'], -1)
        # re-adding a unit -> nil diff
        env.add_units('nova-compute', 1)
        diff = Diff(env, dpl, {}).do_diff()
        self.assertEqual(diff, {})

    def test_diff_config(self):
        dpl = self.get_deployment()
        env = MemoryEnvironment(dpl.name, dpl)
        # Reconfigure the environment to have an explicit change
        # (admin-password), one that differs from the charm
        # default (admin-token), and an option that no longer exists in the
        # charm (obsolete).
        env.set_config(
            'keystone',
            {'admin-password': 'password', 'admin-user': 'root',
             'obsolete': 'for ages'})
        diff = Diff(env, dpl, {}).do_diff()
        mod_keystone = diff['services']['modified']['keystone']
        self.assertEqual(
            mod_keystone['env-config'],
            {'admin-password': 'password', 'admin-user': 'root',
             'obsolete': 'for ages'})
        self.assertEqual(
            mod_keystone['cfg-config'],
            {'admin-password': 'openstack', 'admin-user': 'admin',
             'obsolete': None})

    def test_diff_constraints(self):
        dpl = self.get_deployment()
        env = MemoryEnvironment(dpl.name, dpl)
        env.set_constraints('nova-compute', 'foo=bar')
        diff = Diff(env, dpl, {}).do_diff()
        mod_svc = diff['services']['modified']['nova-compute']
        self.assertTrue(
            mod_svc['env-constraints'] != mod_svc['cfg-constraints'])
        self.assertEqual(mod_svc['env-constraints'], {'foo': 'bar'})

    def test_diff_env_remove_relation(self):
        dpl = self.get_deployment()
        env = MemoryEnvironment(dpl.name, dpl)
        env.remove_relation('cinder:amqp', 'rabbitmq-server:amqp')
        diff = Diff(env, dpl, {}).do_diff()
        self.assertEqual(str(diff['relations']['missing']),
                         '[cinder:amqp <-> rabbitmq-server:amqp]')

    def test_diff_env_service_destroy(self):
        dpl = self.get_deployment()
        env = MemoryEnvironment(dpl.name, dpl)
        env.destroy_service('nova-compute')
        diff = Diff(env, dpl, {}).do_diff()
        self.assertTrue(
            str(diff['relations']['missing'][0]).find('nova-compute') != -1)
        self.assertEqual(
            list(diff['services']['missing'].keys()),
            ['nova-compute'])

    def test_diff_cfg_remove_relation(self):
        dpl = self.get_deployment()
        # need a tmp file to sneak and save deployment YAML
        edited_config_file = os.path.join(self.get_dir(), 'saved.yaml')
        dpl.save(edited_config_file)
        # in memory edit yaml content
        edited_config = yaml.load(open(edited_config_file))
        edited_config["relations"].remove(
            ['nova-compute:amqp', 'rabbitmq-server:amqp'])
        edited_config = yaml.dump({'openstack': edited_config})

        # mock open to inyect edited_config YAML to ConfigStack
        with patch('deployer.config.open', mock_open(read_data=edited_config),
                   create=True):
            edited_dpl = ConfigStack("mocked").get('openstack')

        edited_dpl.repo_path = self.get_dir()
        edited_dpl.fetch_charms()
        edited_dpl.resolve()
        env = MemoryEnvironment(dpl.name, dpl)
        diff = Diff(env, edited_dpl, {}).do_diff()
        self.assertTrue(
            diff['relations']['unknown'],
            '[nova-compute:amqp <-> rabbitmq-server:amqp]')

    def test_diff_cfg_remove_ambiguous_relation(self):
        dpl = self.get_deployment()
        # need a tmp file to sneak and save deployment YAML
        edited_config_file = os.path.join(self.get_dir(), 'saved.yaml')
        dpl.save(edited_config_file)
        # in memory edit yaml content
        edited_config = yaml.load(open(edited_config_file))
        edited_config["relations"].remove(
            ['ceilometer:identity-service', 'keystone:identity-service'])
        edited_config = yaml.dump({'openstack': edited_config})

        # mock open to inyect edited_config YAML to ConfigStack
        with patch('deployer.config.open', mock_open(read_data=edited_config),
                   create=True):
            edited_dpl = ConfigStack("mocked").get('openstack')

        edited_dpl.repo_path = self.get_dir()
        edited_dpl.fetch_charms()
        edited_dpl.resolve()
        env = MemoryEnvironment(dpl.name, dpl)
        diff = Diff(env, edited_dpl, {}).do_diff()
        # verify ceilometer<->keystone present in unknown relations
        self.assertTrue('ceilometer' in str(diff['relations']['unknown']) and
                        'keystone' in str(diff['relations']['unknown']))

    def test_diff_cfg_replace_with_unnamed_relations(self):
        dpl = self.get_deployment()
        # need a tmp file to sneak and save deployment YAML
        edited_config_file = os.path.join(self.get_dir(), 'saved.yaml')
        dpl.save(edited_config_file)
        # in memory edit yaml content
        edited_config = yaml.load(open(edited_config_file))
        # remove ':name' from service endpoint specification
        for relation in edited_config["relations"]:
            relation = [(relation[0] + ":").split(":")[0],
                        (relation[1] + ":").split(":")[0]]
        edited_config = yaml.dump({'openstack': edited_config})

        # mock open to inyect edited_config YAML to ConfigStack
        with patch('deployer.config.open', mock_open(read_data=edited_config),
                   create=True):
            edited_dpl = ConfigStack("mocked").get('openstack')

        edited_dpl.repo_path = self.get_dir()
        edited_dpl.fetch_charms()
        edited_dpl.resolve()
        env = MemoryEnvironment(dpl.name, dpl)
        diff = Diff(env, edited_dpl, {}).do_diff()
        self.assertEqual(diff, {})



from __future__ import absolute_import
from os.path import abspath, isabs, join, dirname

import logging
import os
import tempfile
import shutil

import six
from six.moves.urllib.request import urlopen
from six.moves.urllib.parse import urlparse

from .deployment import Deployment
from .utils import (
    ErrorExit,
    yaml_load,
    path_exists,
    dict_merge,
)


class ConfigStack(object):

    log = logging.getLogger("deployer.config")

    def __init__(self, config_files, cli_series=None):
        self.config_files = config_files
        self.cli_series = cli_series
        self.version = 3
        self.data = {}
        self.yaml = {}
        self.include_dirs = []
        self.urlopen = urlopen
        self.load()

    def _yaml_load(self, config_file):
        if config_file in self.yaml:
            return self.yaml[config_file]

        if urlparse(config_file).scheme:
            response = self.urlopen(config_file)
            if response.getcode() == 200:
                temp = tempfile.NamedTemporaryFile(delete=True)
                shutil.copyfileobj(response, temp)
                temp.flush()
                config_file = temp.name
            else:
                self.log.warning("Could not retrieve %s", config_file)
                raise ErrorExit()

        with open(config_file) as fh:
            try:
                yaml_result = yaml_load(fh.read())
            except Exception as e:
                self.log.warning(
                    "Couldn't load config file @ %r, error: %s:%s",
                    config_file, type(e), e)
                raise

        # Check if this is a v4 bundle.
        services = yaml_result.get('services')
        if (isinstance(services, dict) and
                'services' not in services):
            self.version = 4
            yaml_result = {config_file: yaml_result}

        self.yaml[config_file] = yaml_result

        return self.yaml[config_file]

    def keys(self):
        return sorted(self.data)

    def get(self, key):
        if key not in self.data:
            self.log.warning("Deployment %r not found. Available %s",
                             key, ", ".join(list(self.keys())))
            raise ErrorExit()
        deploy_data = self.data[key]
        if self.version < 4:
            deploy_data = self._resolve_inherited(deploy_data)
        if self.cli_series:
            deploy_data['series'] = self.cli_series
        return Deployment(
            key, deploy_data, self.include_dirs,
            repo_path=os.environ.get("JUJU_REPOSITORY", ""),
            version=self.version)

    def load(self):
        data = {}
        include_dirs = []
        for fp in self._resolve_included():
            if path_exists(fp):
                include_dirs.append(dirname(abspath(fp)))
            d = self._yaml_load(fp)
            data = dict_merge(data, d)
        self.data = data
        for k in ['include-config', 'include-configs']:
            if k in self.data:
                self.data.pop(k)
        self.include_dirs = include_dirs

    def _inherits(self, d):
        parents = d.get('inherits', ())
        if isinstance(parents, six.string_types):
            parents = [parents]
        return parents

    def _resolve_inherited(self, deploy_data):
        if 'inherits' not in deploy_data:
            return deploy_data
        inherits = parents = self._inherits(deploy_data)
        for parent_name in parents:
            parent = self.get(parent_name)
            inherits.extend(self._inherits(parent.data))
            deploy_data = dict_merge(parent.data, deploy_data)

        deploy_data['inherits'] = inherits
        return deploy_data

    def _includes(self, config_file):
        files = [config_file]
        d = self._yaml_load(config_file)

        incs = d.get('include-configs') or d.get('include-config')
        if isinstance(incs, six.string_types):
            inc_fs = [incs]
        else:
            inc_fs = incs

        if inc_fs:
            for inc_f in inc_fs:
                if not isabs(inc_f):
                    inc_f = join(dirname(config_file), inc_f)
                files.extend(self._includes(inc_f))

        return files

    def _resolve_included(self):
        files = []
        [files.extend(self._includes(cf)) for cf in self.config_files]
        return files



from __future__ import absolute_import
from functools import cmp_to_key
import time

from .base import (
    BaseEnvironment,
    NormalizedStatus,
)

from ..utils import (
    ErrorExit,
    get_juju_major_version,
    parse_constraints,
)

from jujuclient.exc import (
    EnvError,
    UnitErrors,
)

from .watchers import (
    raise_on_errors,
    NormalizedDelta,
    WaitForMachineTermination,
    WaitForUnits,
)


class GoEnvironment(BaseEnvironment):

    def __init__(self, name, options=None, endpoint=None):
        self.name = name
        self.options = options
        self.api_endpoint = endpoint
        self.client = None
        self.juju_version = get_juju_major_version()

    @property
    def client_class(self):
        if self.juju_version == 1:
            from jujuclient.juju1.environment import Environment
        else:
            from jujuclient.juju2.environment import Environment
        return Environment

    def add_machine(self, series="", constraints={}):
        """Add a top level machine to the Juju environment.

        Use the given series and constraints.
        Return the machine identifier (e.g. "1").

        series: a string such as 'precise' or 'trusty'.
        constraints: a map of constraints (such as mem, arch, etc.) which
          can be parsed by utils.parse_constraints
        """
        result = self.client.add_machine(
            series=series,
            constraints=parse_constraints(constraints))
        if 'Machine' in result:
            return result['Machine']
        return result['machine']

    def add_unit(self, service_name, machine_spec):
        return self.client.add_unit(service_name, machine_spec)

    def add_units(self, service_name, num_units):
        return self.client.add_units(service_name, num_units)

    def add_relation(self, endpoint_a, endpoint_b):
        return self.client.add_relation(endpoint_a, endpoint_b)

    def close(self):
        if self.client:
            self.client.close()

    def connect(self):
        self.log.debug("Connecting to %s...", self.name)
        self.client = self.client_class.connect(self.name)
        self.log.debug("Connected.")

    def get_config(self, svc_name):
        return self.client.get_config(svc_name)

    def get_constraints(self, svc_name):
        try:
            return self.client.get_constraints(svc_name)
        except EnvError as err:
            if 'constraints do not apply to subordinate' in str(err):
                return {}
            raise

    def expose(self, name):
        return self.client.expose(name)

    def reset(self,
              terminate_machines=False,
              terminate_delay=0,
              timeout=360,
              watch=False,
              force_terminate=False):
        """Destroy/reset the environment."""
        status = self.status()
        destroyed = False
        for s in status.get('services', {}).keys():
            self.log.debug(" Destroying application %s", s)
            self.client.destroy_service(s)
            destroyed = True

        if destroyed:
            # Mark any errors as resolved so destruction can proceed.
            self.resolve_errors()

            if not (terminate_machines and force_terminate):
                # Wait for units to be removed. Only necessary if we're not
                # force-terminating machines.
                self.wait_for_units(timeout, goal_state='removed', watch=watch)

        # The only value to not terminating is keeping the data on the
        # machines around.
        if not terminate_machines:
            self.log.info(
                " warning: juju-core machines are not reusable for units")
            return
        self._terminate_machines(
            status, watch, terminate_delay, force_terminate)

    def _terminate_machines(self, status, watch, terminate_wait, force):
        """Terminate all machines, optionally wait for termination.
        """
        # Terminate machines
        self.log.debug(
            " Terminating machines %s", 'forcefully' if force else '')

        # Don't bother if there are no service unit machines
        if len(status['machines']) == 1 and self.juju_version == 1:
            return

        # containers before machines, container hosts post wait.
        machines = list(status['machines'].keys())

        container_hosts = set()
        containers = set()

        def machine_sort(x, y):
            for ctype in ('lxc', 'lxd', 'kvm'):
                for m in (x, y):
                    if ctype in m:
                        container_hosts.add(m.split('/', 1)[0])
                        containers.add(m)
                        if m == x:
                            return -1
                        if m == y:
                            return 1
            return (x > y) - (x < y)

        machines.sort(key=cmp_to_key(machine_sort))

        for mid in machines:
            self._terminate_machine(mid, container_hosts, force=force)

        if containers:
            watch = self.client.get_watch(120)
            WaitForMachineTermination(
                watch, containers).run(self._delta_event_log)

        for mid in container_hosts:
            self._terminate_machine(mid, force=force)

        if terminate_wait:
            self.log.info("  Waiting for machine termination")
            callback = watch and self._delta_event_log or None
            self.client.wait_for_no_machines(terminate_wait, callback)

    def _terminate_machine(self, mid, container_hosts=(), force=False):
        if mid == "0" and self.juju_version == 1:
            return
        if mid in container_hosts:
            return
        self.log.debug("  Terminating machine %s", mid)
        self.terminate_machine(mid, force=force)

    def _check_timeout(self, etime):
        w_timeout = etime - time.time()
        if w_timeout < 0:
            self.log.error("Timeout reached while resolving errors")
            raise ErrorExit()
        return w_timeout

    def resolve_errors(self, retry_count=0, timeout=600, watch=False, delay=5):
        """Resolve any unit errors in the environment.

        If retry_count is given then the hook execution is reattempted. The
        system will do up to retry_count passes through the system resolving
        errors.

        If retry count is not given, the error is marked resolved permanently.
        """
        etime = time.time() + timeout
        count = 0
        while True:
            error_units = self._get_units_in_error()
            for e_uid in error_units:
                try:
                    self.client.resolved(e_uid, retry=bool(retry_count))
                    self.log.debug("  Resolving error on %s", e_uid)
                except EnvError as err:
                    if 'already resolved' in err:
                        continue

            if not error_units:
                if not count:
                    self.log.debug("  No unit errors found.")
                else:
                    self.log.debug("  No more unit errors found.")
                return

            w_timeout = self._check_timeout(etime)
            if retry_count:
                time.sleep(delay)

            count += 1
            try:
                self.wait_for_units(
                    timeout=int(w_timeout), watch=True,
                    on_errors=raise_on_errors(UnitErrors))
            except UnitErrors as err:
                if retry_count == count:
                    self.log.info(
                        " Retry count %d exhausted, but units in error (%s)",
                        retry_count, " ".join(u['Name'] for u in err.errors))
                    return
            else:
                return

    def set_annotation(self, entity_name, annotation, entity_type='service'):
        """Set an annotation on an entity.

        entity_name: the name of the entity (machine, application, etc.) to
            annotate.
        annotation: a dict of key/value pairs to set on the entity.
        entity_type: the type of entity (machine, application, etc.) to
            annotate.
        """
        return self.client.set_annotation(entity_name, entity_type, annotation)

    def status(self):
        return NormalizedStatus(self.client.status())

    def log_errors(self, errors):
        """Log the given unit errors.

        This can be used in the WaitForUnits error handling machinery, e.g.
        see deployer.watchers.log_on_errors.
        """
        for e in errors:
            if 'StatusInfo' not in e:
                # convert from juju2 format
                e['Name'] = e['name']
                e['MachineId'] = e['machine-id']
                e['Status'] = e['workload-status']['current']
                e['StatusInfo'] = e['workload-status']['message']

        messages = [
            'unit: {Name}: machine: {MachineId} agent-state: {Status} '
            'details: {StatusInfo}'.format(**error) for error in errors
        ]
        self.log.error(
            'The following units had errors:\n   {}'.format(
                '   \n'.join(messages)))

    def wait_for_units(
            self, timeout, goal_state="started", watch=False, services=None,
            on_errors=None):
        """Wait for units to reach a given condition."""
        callback = self._delta_event_log if watch else None
        watcher = self.client.get_watch(timeout)
        WaitForUnits(
            watcher, goal_state=goal_state,
            services=services, on_errors=on_errors).run(callback)

    def _delta_event_log(self, et, ct, d):
        # event type, change type, data
        d = NormalizedDelta(d)
        name = d.get('Name', d.get('Id', 'unknown'))
        state = (
            d.get('workload-status') or
            d.get('Status') or
            d.get('Life') or
            'unknown'
        )
        if et == "relation":
            name = self._format_endpoints(d['Endpoints'])
            state = "created"
            if ct == "remove":
                state = "removed"
        self.log.debug(
            " Delta %s: %s %s:%s", et, name, ct, state)

    def _format_endpoints(self, eps):
        if len(eps) == 1:
            ep = eps.pop()
            return "[%s:%s:%s]" % (
                ep['ServiceName'],
                ep['Relation']['Name'],
                ep['Relation']['Role'])

        return "[%s:%s <-> %s:%s]" % (
            eps[0]['ServiceName'],
            eps[0]['Relation']['Name'],
            eps[1]['ServiceName'],
            eps[1]['Relation']['Name'])



from __future__ import absolute_import
from deployer.utils import parse_constraints
import itertools
from jujuclient.exc import (
    UnitErrors,
    EnvError,
)
from six.moves import range


class MemoryEnvironment(object):
    """ In memory deployment: not all features implemented
        (notably subordinates and their relations).
    """

    def __init__(self, name, deployment):
        """ Two main dicts: _services (return-able as part of status(),
            and _services_data (to hold e.g. config, constraints)
        """
        super(MemoryEnvironment, self).__init__()
        self.name = name
        self._deployment = deployment
        self._services = {}
        self._services_data = {}
        self._relations = {}
        self._do_deploy()

    def add_units(self, svc_name, num):
        """Add units
        """
        next_num = self._services_data[svc_name]['next_unit_num']
        for idx in range(next_num, next_num + num):
            self._services[svc_name]['units'].append(
                '{}/{}'.format(svc_name, idx))
        self._services_data[svc_name]['next_unit_num'] = \
            next_num + num

    def remove_unit(self, unit_name):
        """ Remove a unit by name """
        svc_name = unit_name.split('/')[0]
        units_idx = {unit: idx for idx, unit in
                     enumerate(self._services[svc_name]['units'])}
        try:
            self._services[svc_name]['units'].pop(
                units_idx[unit_name])
        except KeyError:
            raise UnitErrors("Invalid unit name")

    def _get_service(self, svc_name):
        """ Get service by name (as returned by status())
        """
        if svc_name not in self._services:
            raise EnvError("Invalid service name")
        return self._services[svc_name]

    def _remove_relation(self, s_from, r_from, s_to):
        """Low level _remove_relation
        """
        rels = self._relations.get(s_from, {})
        related_svcs = rels.get(r_from, [])
        if s_to in related_svcs:
            related_svcs.remove(s_to)

    def remove_relation(self, src_ep, dst_ep):
        """Remove relation, eps are "service_name:rel_name"
        """
        assert ":" in src_ep and ":" in dst_ep
        (svc0, rel0, svc1, rel1) = src_ep.split(':') + dst_ep.split(':')
        for (s_from, r_from, s_to) in ((svc0, rel0, svc1),
                                       (svc1, rel1, svc0)):
            self._remove_relation(s_from, r_from, s_to)

    def _remove_service_relations(self, svc):
        self._relations.pop(svc, None)
        for _, rels in self._relations.items():
            for _, dst_svcs in rels.items():
                while svc in dst_svcs:
                    dst_svcs.remove(svc)

    def _add_relation(self, s_from, r_from, s_to):
        """Low level _add_relation (only one-way)
        """
        src_rels = self._relations.setdefault(s_from, {})
        src_rels.setdefault(r_from, [])
        src_rels[r_from].append(s_to)

    def add_relation(self, src_ep, dst_ep):
        """Add relation, eps are "service_name:rel_name"
        """
        assert ":" in src_ep and ":" in dst_ep
        (svc0, rel0, svc1, rel1) = src_ep.split(':') + dst_ep.split(':')
        for (s_from, r_from, s_to) in ((svc0, rel0, svc1),
                                       (svc1, rel1, svc0)):
            self._add_relation(s_from, r_from, s_to)

    def destroy_service(self, svc_name):
        """ Destroy a service
        """
        if svc_name not in self._services:
            raise EnvError("Invalid service name")
        self._remove_service_relations(svc_name)
        del self._services[svc_name]

    def close(self):
        """
        """

    def connect(self):
        """
        """

    def set_config(self, svc_name, cfg_dict):
        """ Set service config from passed dict, keeping
            the structure as needed for status() return
        """
        config = self.get_config(svc_name)
        if cfg_dict:
            for cfg_k, cfg_v in cfg_dict.items():
                config_entry = config.setdefault(cfg_k, {})
                config_entry['value'] = cfg_v

    def set_constraints(self, svc_name, constr_str):
        """ Set service constraints from "key=value ..."
            passed string
        """
        constraints = parse_constraints(constr_str) if constr_str else {}
        self._services_data[svc_name]['constraints'] = constraints

    def get_config(self, svc_name):
        """ Return service configs - note its structure:
            config{thename: {'value': thevalue}, ...}
        """
        return self._services_data[svc_name]['config']

    def get_constraints(self, svc_name):
        """ Return service constraints dict
        """
        return self._services_data[svc_name]['constraints']

    def get_cli_status(self):
        pass

    def reset(self):
        pass

    def resolve_errors(self, retry_count=0, timeout=600, watch=False, delay=5):
        pass

    def _do_deploy(self):
        """ Fake deploy: build in-memory representation of the deployed set
            of services from deployment
        """
        self._compile_relations()
        for service in self._deployment.get_services():
            svc_name = service.name
            charm = self._deployment.get_charm_for(svc_name)
            relations = self._relations.setdefault(svc_name, {})
            self._services[svc_name] = {
                'units': [],
                'charm': charm.name,
                'relations': relations,
            }
            self._services_data[svc_name] = {
                'next_unit_num': 0,
                'config': {},
                'constraints': {},
            }
            # XXX: Incomplete relations support: only add units for non-subords
            num_units = 0 if charm.is_subordinate() else service.num_units
            self.add_units(svc_name, num_units)
            self.set_config(svc_name, service.config)
            self.set_constraints(svc_name, service.constraints)

    def _get_charm_if_rels_map(self, svc, req_prov):
        """Return a map of interface names to sets of charm relation names.

        req_prov specifies the relation direction (requires if True,
        otherwise provides).

        e.g. {"rabbitmq": {"amqp", "amqp-nova"}, "http": {"website"}}
        """
        charm = self._deployment.get_charm_for(svc)
        try:
            if req_prov:
                charm_relations = charm.get_requires()
            else:
                charm_relations = charm.get_provides()
        except KeyError:
            return {}
        if_rels_map = {}
        for relations in charm_relations.values():
            for relation in relations:
                interface = relation.get('interface')
                names = if_rels_map.setdefault(interface, set())
                names.add(relation.get('name'))
        return if_rels_map

    def _compile_relations(self):
        """Compile the configured relations into status()-compatible format."""
        for rel in self._deployment.get_relations():
            candidates = []
            # Try both directions.
            for src, dst in (rel[0], rel[1]), (rel[1], rel[0]):
                # src, dst are svcname[:relation_name], e.g.:
                # "rabbitmq" or "rabbitmq:amqp"
                src_svc, src_rel = (src + ':').split(':')[:2]
                dst_svc, dst_rel = (dst + ':').split(':')[:2]

                src_charm_if_rels = self._get_charm_if_rels_map(src_svc, 0)
                dst_charm_if_rels = self._get_charm_if_rels_map(dst_svc, 1)

                # Find all possible relations with matching interfaces
                # between the charms, and filter them down based on any
                # provided relation names.
                for src_charm_if, src_charm_rels in src_charm_if_rels.items():
                    dst_charm_rels = dst_charm_if_rels.get(src_charm_if)
                    if not dst_charm_rels:
                        continue

                    # Filter by name if provided.
                    src_rels = [
                        rel for rel in src_charm_rels
                        if src_rel == rel or not src_rel]
                    dst_rels = [
                        rel for rel in dst_charm_rels
                        if dst_rel == rel or not dst_rel]

                    # All remaining combinations are possibilities.
                    candidates.extend([
                        ((src_svc, src_rel), (dst_svc, dst_rel))
                        for src_rel, dst_rel
                        in itertools.product(src_rels, dst_rels)])

            # Check that we found exactly one possible relation.
            # TODO: Better error handling.
            if not candidates:
                raise Exception(
                    "No relations possible for %s <-> %s" % (src, dst))
            elif len(candidates) > 1:
                raise Exception(
                    "Ambiguous relation %s <-> %s: candidates are %s"
                    % (src, dst,
                       ", ".join([
                           "%s:%s <-> %s:%s" % (
                               src_svc, src_rel, dst_svc, dst_rel)
                           for ((src_svc, src_rel), (dst_svc, dst_rel))
                           in candidates])))

            # Add the identified relation to our state.
            [((src_svc, src_rel), (dst_svc, dst_rel))] = candidates
            self._add_relation(src_svc, src_rel, dst_svc)
            self._add_relation(dst_svc, dst_rel, src_svc)

    def status(self):
        """ Return all services status """
        return {'services': self._services}

    def wait_for_units(self, *args, **kwargs):
        pass
